# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
try:
    from sklearn.neighbors import DistanceMetric as SklearnDistanceMetric
except ImportError:  # pragma: no cover
    SklearnDistanceMetric = None

from .... import opcodes as OperandDef
from ....serialize import KeyField, BoolField
from ....tensor.core import TensorOrder
from ....tensor.indexing import fill_diagonal
from ....tensor.array_utils import as_same_device, device
from ....utils import recursive_tile
from .core import PairwiseDistances


class HaversineDistances(PairwiseDistances):
    _op_type_ = OperandDef.PAIRWISE_HAVERSINE_DISTANCES

    _x = KeyField('x')
    _y = KeyField('y')
    # for test purpose
    _use_sklearn = BoolField('use_sklearn')

    def __init__(self, x=None, y=None, dtype=None, gpu=None,
                 use_sklearn=None, **kw):
        super().__init__(_x=x, _y=y, _dtype=dtype, _gpu=gpu,
                         _use_sklearn=use_sklearn, **kw)
        if self._use_sklearn is None:
            # if not set use_sklearn, will try to use sklearn by default
            self._use_sklearn = True

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def use_sklearn(self):
        return self._use_sklearn

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._x = self._inputs[0]
        self._y = self._inputs[1]

    def __call__(self, X, Y=None):
        X, Y = self.check_pairwise_arrays(X, Y)
        if self._y is None:
            self._y = Y

        if X.shape[1] != 2 or Y.shape[1] != 2:
            raise ValueError('Haversine distance only valid in 2 dimensions')
        if X.issparse() or Y.issparse():
            raise TypeError('Haversine distance requires inputs dense')

        return self.new_tensor([X, Y], shape=(X.shape[0], Y.shape[0]),
                               order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        x, y = op.x, op.y
        y_is_x = y is x

        if len(x.chunks) == 1 and len(y.chunks) == 1:
            return cls._tile_one_chunk(op)

        x, y = cls._rechunk_cols_into_one(x, y)
        ret, = cls._tile_chunks(op, x, y)
        if y_is_x:
            fill_diagonal(ret, 0)
        return [recursive_tile(ret)]

    @classmethod
    def execute(cls, ctx, op):
        (x, y), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            if xp is np and op.use_sklearn and SklearnDistanceMetric is not None:
                # CPU and sklearn installed, delegate computation to sklearn
                d = SklearnDistanceMetric.get_metric('haversine').pairwise(x, y)
            else:
                # try to leverage xp(np, cp) to perform computation
                sin_0 = xp.sin(0.5 * (x[:, [0]] - y[:, 0]))
                sin_1 = xp.sin(0.5 * (x[:, [1]] - y[:, 1]))
                d = 2 * xp.arcsin(xp.sqrt(
                    sin_0 * sin_0 + xp.cos(x[:, [0]]) * xp.cos(y[:, 0]) * sin_1 * sin_1))

            ctx[op.outputs[0].key] = d


def haversine_distances(X, Y=None):
    """Compute the Haversine distance between samples in X and Y

    The Haversine (or great circle) distance is the angular distance between
    two points on the surface of a sphere. The first distance of each point is
    assumed to be the latitude, the second is the longitude, given in radians.
    The dimension of the data must be 2.

    .. math::
       D(x, y) = 2\\arcsin[\\sqrt{\\sin^2((x1 - y1) / 2)
                                + \\cos(x1)\\cos(y1)\\sin^2((x2 - y2) / 2)}]

    Parameters
    ----------
    X : array_like, shape (n_samples_1, 2)

    Y : array_like, shape (n_samples_2, 2), optional

    Returns
    -------
    distance : {Tensor}, shape (n_samples_1, n_samples_2)

    Notes
    -----
    As the Earth is nearly spherical, the haversine formula provides a good
    approximation of the distance between two points of the Earth surface, with
    a less than 1% error on average.

    Examples
    --------
    We want to calculate the distance between the Ezeiza Airport
    (Buenos Aires, Argentina) and the Charles de Gaulle Airport (Paris, France)

    >>> from mars.learn.metrics.pairwise import haversine_distances
    >>> bsas = [-34.83333, -58.5166646]
    >>> paris = [49.0083899664, 2.53844117956]
    >>> result = haversine_distances([bsas, paris])
    >>> (result * 6371000/1000).execute()  # multiply by Earth radius to get kilometers
    array([[    0.        , 11279.45379464],
           [11279.45379464,     0.        ]])
    """
    op = HaversineDistances(x=X, y=Y, dtype=np.dtype(np.float64))
    return op(X, Y=Y)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from abc import ABC, abstractmethod
from urllib.parse import urlparse
from typing import Union, List, Dict, Tuple, BinaryIO, TextIO, Iterator

from ...utils import stringify_path

path_type = Union[str, os.PathLike]


class FileSystem(ABC):
    """
    Abstract filesystem interface
    """

    @abstractmethod
    def cat(self, path: path_type) -> bytes:
        """
        Return contents of file as a bytes object

        Parameters
        ----------
        path : str or path-like
            File path to read content from.

        Returns
        -------
        contents : bytes
        """

    @abstractmethod
    def ls(self, path: path_type) -> List[path_type]:
        """
        Return list of file paths

        Returns
        -------
        paths : list
        """

    @abstractmethod
    def delete(self,
               path: path_type,
               recursive: bool = False):
        """
        Delete the indicated file or directory

        Parameters
        ----------
        path : str
        recursive : bool, default False
            If True, also delete child paths for directories
        """

    def disk_usage(self, path: path_type) -> int:
        """
        Compute bytes used by all contents under indicated path in file tree

        Parameters
        ----------
        path : string
            Can be a file path or directory

        Returns
        -------
        usage : int
        """
        path = stringify_path(path)
        path_info = self.stat(path)
        if path_info['type'] == 'file':
            return path_info['size']

        total = 0
        for root, directories, files in self.walk(path):
            for child_path in files:
                abspath = self.path_join(root, child_path)
                total += self.stat(abspath)['size']

        return total

    def path_join(self, *args):
        return self.pathsep.join(args)

    def path_split(self, path):
        return path.rsplit(self.pathsep, 1)

    @abstractmethod
    def stat(self, path: path_type) -> Dict:
        """
        Information about a filesystem entry.

        Returns
        -------
        stat : dict
        """

    def rm(self,
           path: path_type,
           recursive: bool = False):
        """
        Alias for FileSystem.delete
        """
        return self.delete(path, recursive=recursive)

    def mv(self, path, new_path):
        """
        Alias for FileSystem.rename
        """
        return self.rename(path, new_path)

    @abstractmethod
    def rename(self,
               path: path_type,
               new_path: path_type):
        """
        Rename file, like UNIX mv command

        Parameters
        ----------
        path : string
            Path to alter
        new_path : string
            Path to move to
        """

    @abstractmethod
    def mkdir(self,
              path: path_type,
              create_parents: bool = True):
        """
        Create a directory.

        Parameters
        ----------
        path : str
            Path to the directory.
        create_parents : bool, default True
            If the parent directories don't exists create them as well.
        """

    @abstractmethod
    def exists(self, path: path_type):
        """
        Return True if path exists.

        Parameters
        ----------
        path : str
            Path to check.
        """

    @abstractmethod
    def isdir(self, path: path_type) -> bool:
        """
        Return True if path is a directory.

        Parameters
        ----------
        path : str
            Path to check.
        """

    @abstractmethod
    def isfile(self, path: path_type) -> bool:
        """
        Return True if path is a file.

        Parameters
        ----------
        path : str
            Path to check.
        """

    @abstractmethod
    def _isfilestore(self) -> bool:
        """
        Returns True if this FileSystem is a unix-style file store with
        directories.
        """

    @abstractmethod
    def open(self,
             path: path_type,
             mode: str = 'rb') -> Union[BinaryIO, TextIO]:
        """
        Open file for reading or writing.
        """

    @abstractmethod
    def walk(self, path: path_type) -> Iterator[Tuple[str, List[str], List[str]]]:
        """
        Directory tree generator.

        Parameters
        ----------
        path : str

        Returns
        -------
        generator
        """

    @abstractmethod
    def glob(self,
             path: path_type,
             recursive: bool = False) -> List[path_type]:
        """
        Return a list of paths matching a pathname pattern.

        Parameters
        ----------
        path : str
            Pattern may contain simple shell-style wildcards
        recursive : bool
            If recursive is true, the pattern '**' will match any files and
            zero or more directories and subdirectories.

        Returns
        -------
        paths : List
        """

    @property
    def pathsep(self) -> str:
        return '/'

    @staticmethod
    def parse_from_path(uri: str):
        parsed_uri = urlparse(uri)
        options = dict()
        options['host'] = parsed_uri.netloc.rsplit("@", 1)[-1].rsplit(":", 1)[0]
        if parsed_uri.port:
            options["port"] = parsed_uri.port
        if parsed_uri.username:
            options["user"] = parsed_uri.username
        if parsed_uri.password:
            options["password"] = parsed_uri.password
        return options



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import time
import logging
from collections import defaultdict

from .utils import WorkerActor, ExpMeanHolder
from .. import resource
from ..config import options
from ..node_info import gather_node_info

logger = logging.getLogger(__name__)


class StatusReporterActor(WorkerActor):
    def __init__(self, endpoint, with_gpu=True):
        super().__init__()
        self._endpoint = endpoint
        self._with_gpu = with_gpu

        self._upload_status = False

        self._status_ref = None
        self._resource_ref = None

    def post_create(self):
        from ..scheduler import ResourceActor

        super().post_create()
        self._status_ref = self.ctx.actor_ref(StatusActor.default_uid())
        self._resource_ref = self.get_actor_ref(ResourceActor.default_uid())
        self.ref().collect_status(_tell=True)

    def enable_status_upload(self):
        self._upload_status = True

    def collect_status(self):
        """
        Collect worker status and write to kvstore
        """
        meta_dict = dict()
        try:
            if not self._upload_status:
                return

            cpu_percent = resource.cpu_percent()
            disk_io = resource.disk_io_usage()
            net_io = resource.net_io_usage()
            if cpu_percent is None or disk_io is None or net_io is None:
                return
            hw_metrics = dict()
            hw_metrics['cpu'] = max(0.0, resource.cpu_count() - cpu_percent / 100.0)
            hw_metrics['cpu_used'] = cpu_percent / 100.0
            hw_metrics['cpu_total'] = resource.cpu_count()

            cuda_info = resource.cuda_info() if self._with_gpu else None
            if cuda_info:
                hw_metrics['cuda'] = cuda_info.gpu_count
                hw_metrics['cuda_total'] = cuda_info.gpu_count

            hw_metrics['disk_read'] = disk_io[0]
            hw_metrics['disk_write'] = disk_io[1]
            hw_metrics['net_receive'] = net_io[0]
            hw_metrics['net_send'] = net_io[1]

            iowait = resource.iowait()
            if iowait is not None:
                hw_metrics['iowait'] = iowait

            mem_stats = resource.virtual_memory()
            hw_metrics['memory'] = int(mem_stats.available)

            hw_metrics['memory_used'] = int(mem_stats.used)
            hw_metrics['memory_total'] = int(mem_stats.total)

            cache_allocations = self._status_ref.get_cache_allocations()
            cache_total = cache_allocations.get('total', 0)
            hw_metrics['cached_total'] = int(cache_total)
            hw_metrics['cached_hold'] = int(cache_allocations.get('hold', 0))

            mem_quota_allocations = self._status_ref.get_mem_quota_allocations()
            mem_quota_total = mem_quota_allocations.get('total', 0)
            mem_quota_allocated = mem_quota_allocations.get('allocated', 0)
            if not mem_quota_allocations:
                hw_metrics['mem_quota'] = hw_metrics['memory']
                hw_metrics['mem_quota_used'] = hw_metrics['memory_used']
                hw_metrics['mem_quota_total'] = hw_metrics['memory_total']
                hw_metrics['mem_quota_hold'] = 0
            else:
                hw_metrics['mem_quota'] = int(mem_quota_total - mem_quota_allocated) or hw_metrics['memory']
                hw_metrics['mem_quota_used'] = int(mem_quota_allocated)
                hw_metrics['mem_quota_total'] = int(mem_quota_total)
                hw_metrics['mem_quota_hold'] = int(mem_quota_allocations.get('hold', 0))

            if options.worker.spill_directory:
                if isinstance(options.worker.spill_directory, str):
                    spill_dirs = options.worker.spill_directory.split(':')
                else:
                    spill_dirs = options.worker.spill_directory
                if spill_dirs and 'disk_stats' not in hw_metrics:
                    hw_metrics['disk_stats'] = dict()
                disk_stats = hw_metrics['disk_stats']

                agg_disk_used = agg_disk_total = 0.0
                agg_inode_used = agg_inode_total = 0
                for spill_dir in spill_dirs:
                    if not os.path.exists(spill_dir):
                        continue
                    if spill_dir not in disk_stats:
                        disk_stats[spill_dir] = dict()

                    disk_usage = resource.disk_usage(spill_dir)
                    disk_stats[spill_dir]['disk_total'] = disk_usage.total
                    agg_disk_total += disk_usage.total
                    disk_stats[spill_dir]['disk_used'] = disk_usage.used
                    agg_disk_used += disk_usage.used

                    vfs_stat = os.statvfs(spill_dir)
                    disk_stats[spill_dir]['inode_total'] = vfs_stat.f_files
                    agg_inode_total += vfs_stat.f_files
                    disk_stats[spill_dir]['inode_used'] = vfs_stat.f_files - vfs_stat.f_favail
                    agg_inode_used += vfs_stat.f_files - vfs_stat.f_favail
                hw_metrics['disk_used'] = agg_disk_used
                hw_metrics['disk_total'] = agg_disk_total
                hw_metrics['inode_used'] = agg_inode_used
                hw_metrics['inode_total'] = agg_inode_total

            cuda_card_stats = resource.cuda_card_stats() if self._with_gpu else None
            if cuda_card_stats:
                hw_metrics['cuda_stats'] = [dict(
                    product_name=stat.product_name,
                    gpu_usage=stat.gpu_usage,
                    temperature=stat.temperature,
                    fb_memory_total=stat.fb_mem_info.total,
                    fb_memory_used=stat.fb_mem_info.used,
                ) for stat in cuda_card_stats]

            meta_dict = dict()
            meta_dict['hardware'] = hw_metrics
            meta_dict['update_time'] = time.time()
            meta_dict['stats'] = dict()
            meta_dict['slots'] = dict()

            status_data = self._status_ref.get_stats()
            for k, v in status_data.items():
                meta_dict['stats'][k] = v

            slots_data = self._status_ref.get_slots()
            for k, v in slots_data.items():
                meta_dict['slots'][k] = v

            meta_dict['progress'] = self._status_ref.get_progress()
            meta_dict['details'] = gather_node_info()

            if options.vineyard.enabled:  # pragma: no cover
                import vineyard
                client = vineyard.connect(options.vineyard.socket)
                meta_dict['vineyard'] = {'instance_id': client.instance_id}

            self._resource_ref.set_worker_meta(self._endpoint, meta_dict)
        except Exception as ex:
            logger.error('Failed to save status: %s. repr(meta_dict)=%r', str(ex), meta_dict)
        finally:
            self.ref().collect_status(_tell=True, _delay=1)


class StatusActor(WorkerActor):
    def __init__(self, endpoint, with_gpu=True):
        super().__init__()
        self._speed_holders = defaultdict(ExpMeanHolder)
        self._endpoint = endpoint
        self._with_gpu = with_gpu
        self._reporter_ref = None
        self._stats = dict()
        self._slots = dict()
        self._progress = dict()

        self._mem_quota_allocations = {}
        self._cache_allocations = {}

    def post_create(self):
        super().post_create()
        self._reporter_ref = self.ctx.create_actor(
            StatusReporterActor, self._endpoint, with_gpu=self._with_gpu,
            uid=StatusReporterActor.default_uid())

    def pre_destroy(self):
        self.ctx.destroy_actor(self._reporter_ref)

    def enable_status_upload(self):
        self._reporter_ref.enable_status_upload(_tell=True)

    def get_stats(self, items=None):
        if not items:
            return self._stats
        else:
            return dict((k, self._stats[k]) for k in items if k in self._stats)

    def update_stats(self, update_dict):
        self._stats.update(update_dict)

    def get_slots(self):
        return copy.deepcopy(self._slots)

    def update_slots(self, slots_dict):
        self._slots.update(slots_dict)

    def get_progress(self):
        return copy.deepcopy(self._progress)

    def update_progress(self, session_id, graph_key, op_name, state):
        """
        Update statuses of operands
        :param session_id: session id
        :param graph_key: graph key
        :param op_name: operand name
        :param state: operand execution state
        """
        session_id = str(session_id)
        graph_key = str(graph_key)

        try:
            session_dict = self._progress[session_id]
        except KeyError:
            session_dict = self._progress[session_id] = dict()
        try:
            graph_dict = session_dict[graph_key]
        except KeyError:
            graph_dict = session_dict[graph_key] = dict()
        graph_dict.update(dict(operands=op_name, stage=state.name))

    def remove_progress(self, session_id, graph_key):
        try:
            del self._progress[session_id][graph_key]
            if not self._progress[session_id]:
                del self._progress[session_id]
        except KeyError:
            pass

    def update_mean_stats(self, item, value):
        """
        Update sequence statistics. Moment statistics will be computed automatically
        :param item: statistics item
        :param value: statistics value
        """
        stats_item = self._speed_holders[item]
        stats_item.put(value)
        self._stats[item] = dict(count=stats_item.count(), mean=stats_item.mean(),
                                 std=stats_item.std())

    def set_cache_allocations(self, value):
        self._cache_allocations = value

    def get_cache_allocations(self):
        return self._cache_allocations

    def set_mem_quota_allocations(self, value):
        self._mem_quota_allocations = value

    def get_mem_quota_allocations(self):
        return self._mem_quota_allocations



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...operands import FuseChunkMixin
from ..operands import TensorFuse, TensorOperandMixin


class TensorFuseChunkMixin(FuseChunkMixin, TensorOperandMixin):
    __slots__ = ()


class TensorFuseChunk(TensorFuse, TensorFuseChunkMixin):
    def __init__(self, dtype=None, **kw):
        super().__init__(_dtype=dtype, **kw)


def estimate_fuse_size(ctx, op):
    from ...graph import DAG
    from ...executor import Executor
    from ...utils import build_fetch_chunk

    chunk = op.outputs[0]
    dag = DAG()
    size_ctx = dict()
    keys = set(c.key for c in chunk.composed)
    for c in chunk.composed:
        dag.add_node(c)
        for inp in c.inputs:
            if inp.key not in keys:
                size_ctx[inp.key] = ctx[inp.key]
                inp = build_fetch_chunk(inp).data
            if inp not in dag:
                dag.add_node(inp)
            dag.add_edge(inp, c)

    executor = Executor(storage=size_ctx)
    output_keys = [o.key for o in op.outputs]
    results = executor.execute_graph(dag, output_keys, mock=True, no_intermediate=True)
    ctx.update(zip(output_keys, results))

    # update with the maximal memory cost during the whole execution
    total_mem = sum(ctx[key][1] for key in output_keys)
    if total_mem:
        for key in output_keys:
            r = ctx[key]
            ctx[key] = (r[0], max(r[1], r[1] * executor.mock_max_memory // total_mem))



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...operands import OperandStage
from ...serialize import AnyField, Int32Field, ListField, StringField, ValueType
from ..operands import DataFrameOperand, DataFrameOperandMixin, SERIES_TYPE, SERIES_CHUNK_TYPE
from ..utils import build_df, build_series, parse_index


class DataFrameReplace(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.REPLACE

    _to_replace = AnyField('to_replace')
    _value = AnyField('value')
    _limit = Int32Field('limit')
    _regex = AnyField('regex')
    _method = StringField('method')

    _fill_chunks = ListField('fill_chunks', ValueType.key)

    def __init__(self, to_replace=None, value=None, limit=None, regex=None, method=None,
                 fill_chunks=None, **kw):
        super().__init__(_to_replace=to_replace, _value=value, _limit=limit, _regex=regex,
                         _method=method, _fill_chunks=fill_chunks, **kw)

    @property
    def to_replace(self):
        return self._to_replace

    @property
    def value(self):
        return self._value

    @property
    def limit(self):
        return self._limit

    @property
    def regex(self):
        return self._regex

    @property
    def method(self):
        return self._method

    @property
    def fill_chunks(self):
        return self._fill_chunks

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        input_iter = iter(inputs)
        next(input_iter)
        if isinstance(self.to_replace, (SERIES_TYPE, SERIES_CHUNK_TYPE)):
            self._to_replace = next(input_iter)
        if isinstance(self.value, (SERIES_TYPE, SERIES_CHUNK_TYPE)):
            self._value = next(input_iter)
        self._fill_chunks = list(input_iter)

    def __call__(self, df_or_series):
        inputs = [df_or_series]
        mock_obj = build_df(df_or_series) \
            if df_or_series.ndim == 2 else build_series(df_or_series)

        if isinstance(self.to_replace, SERIES_TYPE):
            mock_to_replace = build_series(self.to_replace)
            inputs.append(self.to_replace)
        else:
            mock_to_replace = self.to_replace

        if isinstance(self.value, SERIES_TYPE):
            mock_value = build_series(self.value)
            inputs.append(self.value)
        else:
            mock_value = self.value

        mock_result = mock_obj.replace(
            mock_to_replace, mock_value, regex=self.regex, method=self.method)

        if df_or_series.ndim == 2:
            return self.new_dataframe(
                inputs, shape=df_or_series.shape, dtypes=mock_result.dtypes,
                index_value=df_or_series.index_value, columns_value=df_or_series.columns_value)
        else:
            return self.new_series(
                inputs, shape=df_or_series.shape, dtype=mock_result.dtype,
                index_value=df_or_series.index_value)

    @classmethod
    def _build_result_chunk(cls, op: 'DataFrameReplace', in_chunks, with_fill=False,
                            stage=None):
        in_obj = op.inputs[0]
        out_obj = op.outputs[0]
        in_chunk = in_chunks[0]

        kw = in_chunk.params
        new_shape = list(in_chunk.shape)
        if with_fill:
            new_shape[0] = 1

        if in_obj.ndim == 2:
            new_dtypes = out_obj.dtypes[in_chunk.dtypes.index]
            kw.update(dict(
                dtypes=new_dtypes,
                shape=tuple(new_shape),
                column_values=parse_index(new_dtypes.index),
            ))
        else:
            kw.update(dict(
                dtype=out_obj.dtype,
                shape=tuple(new_shape),
            ))

        new_op = op.copy().reset_key()
        new_op._stage = stage
        return new_op.new_chunk(in_chunks, **kw)

    @classmethod
    def tile(cls, op: 'DataFrameReplace'):
        in_obj = op.inputs[0]
        out_obj = op.outputs[0]

        chunk_inputs_ex = []
        tileable_inputs_ex = []
        to_replace = op.to_replace
        if isinstance(to_replace, SERIES_TYPE):
            to_replace = to_replace.rechunk((to_replace.shape[0],))._inplace_tile()
            chunk_inputs_ex.append(to_replace.chunks[0])
            tileable_inputs_ex.append(to_replace)
        value = op.value
        if isinstance(value, SERIES_TYPE):
            value = value.rechunk((value.shape[0],))._inplace_tile()
            chunk_inputs_ex.append(value.chunks[0])
            tileable_inputs_ex.append(value)

        # fill methods only available when `to_replace` is a scalar, list or tuple
        # and `value` is None.
        with_fill = op.value is None and not isinstance(op.to_replace, dict) \
            and op.method is not None

        chunks = []
        if not with_fill:
            for in_chunk in in_obj.chunks:
                inputs = [in_chunk] + chunk_inputs_ex
                chunks.append(cls._build_result_chunk(
                    op, inputs, with_fill, OperandStage.map if with_fill else None))
        else:
            map_array = np.empty(out_obj.shape, dtype=object)
            for in_chunk in in_obj.chunks:
                inputs = [in_chunk] + chunk_inputs_ex
                map_array[in_chunk.index] = cls._build_result_chunk(
                    op, inputs, with_fill, OperandStage.map if with_fill else None)

            for in_chunk in in_obj.chunks:
                if op.method in ('pad', 'ffill'):
                    slc = slice(0, in_chunk.index[0])
                else:
                    slc = slice(in_chunk.index[0] + 1, in_obj.chunk_shape[0])

                if in_chunk.ndim == 2:
                    append_chunks = list(map_array[slc, in_chunk.index[1]])
                else:
                    append_chunks = list(map_array[slc])

                inputs = [in_chunk] + chunk_inputs_ex + append_chunks
                chunks.append(cls._build_result_chunk(op, inputs, False, OperandStage.combine))

        inputs = [in_obj] + tileable_inputs_ex
        new_op = op.copy().reset_key()
        return new_op.new_tileables(inputs, chunks=chunks, nsplits=in_obj.nsplits, **out_obj.params)

    @classmethod
    def execute(cls, ctx, op: 'DataFrameReplace'):
        in_data = ctx[op.inputs[0].key]
        to_replace = op.to_replace
        if isinstance(to_replace, SERIES_CHUNK_TYPE):
            to_replace = ctx[to_replace.key]
        value = op.value
        if isinstance(value, SERIES_CHUNK_TYPE):
            value = ctx[value.key]

        if not op.fill_chunks:
            concat_data = in_data
        else:
            to_concat = [ctx[c.key] for c in op.fill_chunks]
            if op.method in ('pad', 'ffill'):
                to_concat += [in_data]
            else:
                to_concat = [in_data] + to_concat
            concat_data = pd.concat(to_concat)

        result = concat_data.replace(
            to_replace, value, regex=op.regex, method=op.method, limit=op.limit)
        del concat_data

        if op.stage == OperandStage.map:
            to_slice = op.outputs[0].shape[0]
            if op.method in ('pad', 'ffill'):
                result = result.iloc[-to_slice:]
            else:
                result = result.iloc[:to_slice]
        else:
            to_remove = len(result) - len(in_data)
            if to_remove > 0:
                if op.method in ('pad', 'ffill'):
                    result = result.iloc[to_remove:]
                else:
                    result = result.iloc[:-to_remove]
        ctx[op.outputs[0].key] = result


_fun_doc = """
Replace values given in `to_replace` with `value`.

Values of the #obj_type# are replaced with other values dynamically.
This differs from updating with ``.loc`` or ``.iloc``, which require
you to specify a location to update with some value.

Parameters
----------
to_replace : str, regex, list, dict, Series, int, float, or None
    How to find the values that will be replaced.

    * numeric, str or regex:

        - numeric: numeric values equal to `to_replace` will be
          replaced with `value`
        - str: string exactly matching `to_replace` will be replaced
          with `value`
        - regex: regexs matching `to_replace` will be replaced with
          `value`

    * list of str, regex, or numeric:

        - First, if `to_replace` and `value` are both lists, they
          **must** be the same length.
        - Second, if ``regex=True`` then all of the strings in **both**
          lists will be interpreted as regexs otherwise they will match
          directly. This doesn't matter much for `value` since there
          are only a few possible substitution regexes you can use.
        - str, regex and numeric rules apply as above.

    * dict:

        - Dicts can be used to specify different replacement values
          for different existing values. For example,
          ``{'a': 'b', 'y': 'z'}`` replaces the value 'a' with 'b' and
          'y' with 'z'. To use a dict in this way the `value`
          parameter should be `None`.
        - For a DataFrame a dict can specify that different values
          should be replaced in different columns. For example,
          ``{'a': 1, 'b': 'z'}`` looks for the value 1 in column 'a'
          and the value 'z' in column 'b' and replaces these values
          with whatever is specified in `value`. The `value` parameter
          should not be ``None`` in this case. You can treat this as a
          special case of passing two lists except that you are
          specifying the column to search in.
        - For a DataFrame nested dictionaries, e.g.,
          ``{'a': {'b': np.nan}}``, are read as follows: look in column
          'a' for the value 'b' and replace it with NaN. The `value`
          parameter should be ``None`` to use a nested dict in this
          way. You can nest regular expressions as well. Note that
          column names (the top-level dictionary keys in a nested
          dictionary) **cannot** be regular expressions.

    * None:

        - This means that the `regex` argument must be a string,
          compiled regular expression, or list, dict, ndarray or
          Series of such elements. If `value` is also ``None`` then
          this **must** be a nested dictionary or Series.

    See the examples section for examples of each of these.
value : scalar, dict, list, str, regex, default None
    Value to replace any values matching `to_replace` with.
    For a DataFrame a dict of values can be used to specify which
    value to use for each column (columns not in the dict will not be
    filled). Regular expressions, strings and lists or dicts of such
    objects are also allowed.
inplace : bool, default False
    If True, in place. Note: this will modify any
    other views on this object (e.g. a column from a DataFrame).
    Returns the caller if this is True.
limit : int, default None
    Maximum size gap to forward or backward fill.
regex : bool or same types as `to_replace`, default False
    Whether to interpret `to_replace` and/or `value` as regular
    expressions. If this is ``True`` then `to_replace` *must* be a
    string. Alternatively, this could be a regular expression or a
    list, dict, or array of regular expressions in which case
    `to_replace` must be ``None``.
method : {'pad', 'ffill', 'bfill', `None`}
    The method to use when for replacement, when `to_replace` is a
    scalar, list or tuple and `value` is ``None``.

Returns
-------
#obj_type#
    Object after replacement.

Raises
------
AssertionError
    * If `regex` is not a ``bool`` and `to_replace` is not
      ``None``.
TypeError
    * If `to_replace` is a ``dict`` and `value` is not a ``list``,
      ``dict``, ``ndarray``, or ``Series``
    * If `to_replace` is ``None`` and `regex` is not compilable
      into a regular expression or is a list, dict, ndarray, or
      Series.
    * When replacing multiple ``bool`` or ``datetime64`` objects and
      the arguments to `to_replace` does not match the type of the
      value being replaced
ValueError
    * If a ``list`` or an ``ndarray`` is passed to `to_replace` and
      `value` but they are not the same length.

See Also
--------
#obj_type#.fillna : Fill NA values.
#obj_type#.where : Replace values based on boolean condition.
Series.str.replace : Simple string replacement.

Notes
-----
* Regex substitution is performed under the hood with ``re.sub``. The
  rules for substitution for ``re.sub`` are the same.
* Regular expressions will only substitute on strings, meaning you
  cannot provide, for example, a regular expression matching floating
  point numbers and expect the columns in your frame that have a
  numeric dtype to be matched. However, if those floating point
  numbers *are* strings, then you can do this.
* This method has *a lot* of options. You are encouraged to experiment
  and play with this method to gain intuition about how it works.
* When dict is used as the `to_replace` value, it is like
  key(s) in the dict are the to_replace part and
  value(s) in the dict are the value parameter.

Examples
--------

**Scalar `to_replace` and `value`**

>>> import mars.tensor as mt
>>> import mars.dataframe as md
>>> s = md.Series([0, 1, 2, 3, 4])
>>> s.replace(0, 5).execute()
0    5
1    1
2    2
3    3
4    4
dtype: int64

>>> df = md.DataFrame({'A': [0, 1, 2, 3, 4],
...                    'B': [5, 6, 7, 8, 9],
...                    'C': ['a', 'b', 'c', 'd', 'e']})
>>> df.replace(0, 5).execute()
   A  B  C
0  5  5  a
1  1  6  b
2  2  7  c
3  3  8  d
4  4  9  e

**List-like `to_replace`**

>>> df.replace([0, 1, 2, 3], 4).execute()
   A  B  C
0  4  5  a
1  4  6  b
2  4  7  c
3  4  8  d
4  4  9  e

>>> df.replace([0, 1, 2, 3], [4, 3, 2, 1]).execute()
   A  B  C
0  4  5  a
1  3  6  b
2  2  7  c
3  1  8  d
4  4  9  e

>>> s.replace([1, 2], method='bfill').execute()
0    0
1    3
2    3
3    3
4    4
dtype: int64

**dict-like `to_replace`**

>>> df.replace({0: 10, 1: 100}).execute()
     A  B  C
0   10  5  a
1  100  6  b
2    2  7  c
3    3  8  d
4    4  9  e

>>> df.replace({'A': 0, 'B': 5}, 100).execute()
     A    B  C
0  100  100  a
1    1    6  b
2    2    7  c
3    3    8  d
4    4    9  e

>>> df.replace({'A': {0: 100, 4: 400}}).execute()
     A  B  C
0  100  5  a
1    1  6  b
2    2  7  c
3    3  8  d
4  400  9  e

**Regular expression `to_replace`**

>>> df = md.DataFrame({'A': ['bat', 'foo', 'bait'],
...                    'B': ['abc', 'bar', 'xyz']})
>>> df.replace(to_replace=r'^ba.$', value='new', regex=True).execute()
      A    B
0   new  abc
1   foo  new
2  bait  xyz

>>> df.replace({'A': r'^ba.$'}, {'A': 'new'}, regex=True).execute()
      A    B
0   new  abc
1   foo  bar
2  bait  xyz

>>> df.replace(regex=r'^ba.$', value='new').execute()
      A    B
0   new  abc
1   foo  new
2  bait  xyz

>>> df.replace(regex={r'^ba.$': 'new', 'foo': 'xyz'}).execute()
      A    B
0   new  abc
1   xyz  new
2  bait  xyz

>>> df.replace(regex=[r'^ba.$', 'foo'], value='new').execute()
      A    B
0   new  abc
1   new  new
2  bait  xyz

Note that when replacing multiple ``bool`` or ``datetime64`` objects,
the data types in the `to_replace` parameter must match the data
type of the value being replaced:

>>> df = md.DataFrame({'A': [True, False, True],
...                    'B': [False, True, False]})
>>> df.replace({'a string': 'new value', True: False})  # raises.execute()
Traceback (most recent call last):
    ....execute()
TypeError: Cannot compare types 'ndarray(dtype=bool)' and 'str'

This raises a ``TypeError`` because one of the ``dict`` keys is not of
the correct type for replacement.

Compare the behavior of ``s.replace({'a': None})`` and
``s.replace('a', None)`` to understand the peculiarities
of the `to_replace` parameter:

>>> s = md.Series([10, 'a', 'a', 'b', 'a'])

When one uses a dict as the `to_replace` value, it is like the
value(s) in the dict are equal to the `value` parameter.
``s.replace({'a': None})`` is equivalent to
``s.replace(to_replace={'a': None}, value=None, method=None)``:

>>> s.replace({'a': None}).execute()
0      10
1    None
2    None
3       b
4    None
dtype: object

When ``value=None`` and `to_replace` is a scalar, list or
tuple, `replace` uses the method parameter (default 'pad') to do the
replacement. So this is why the 'a' values are being replaced by 10
in rows 1 and 2 and 'b' in row 4 in this case.
The command ``s.replace('a', None)`` is actually equivalent to
``s.replace(to_replace='a', value=None, method='pad')``:

>>> s.replace('a', None).execute()
0    10
1    10
2    10
3     b
4     b
dtype: object
"""


def _replace(df_or_series, to_replace=None, value=None, inplace=False, limit=None,
             regex=False, method='pad'):
    if not isinstance(to_replace, dict) and value is None and limit is not None:
        raise NotImplementedError('fill with limit not supported when value is None')

    if not isinstance(regex, bool):
        to_replace = regex
        regex = True
    op = DataFrameReplace(to_replace=to_replace, value=value, limit=limit, regex=regex,
                          method=method)
    ret = op(df_or_series)
    if inplace:
        df_or_series.data = ret.data
    else:
        return ret


def df_replace(df, to_replace=None, value=None, inplace=False, limit=None,
               regex=False, method='pad'):
    return _replace(df, to_replace=to_replace, value=value, inplace=inplace, limit=limit,
                    regex=regex, method=method)


def series_replace(series, to_replace=None, value=None, inplace=False, limit=None,
                   regex=False, method='pad'):
    return _replace(series, to_replace=to_replace, value=value, inplace=inplace,
                    limit=limit, regex=regex, method=method)


df_replace.__doc__ = _fun_doc.replace('#obj_type#', 'DataFrame')
series_replace.__doc__ = _fun_doc.replace('#obj_type#', 'Series')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
import copy
from collections import OrderedDict
from collections.abc import Iterable

from .._utils cimport to_binary
from ..lib.mmh3 import hash as mmh_hash

cdef dict _serializable_registry = dict()


cdef class Identity:
    def __init__(self, tp=None):
        if tp is None:
            tp = PrimitiveType.unicode
        self.type = tp
        self.name = 'id'

    def __call__(self, tp):
        return Identity(tp)


cdef class List:
    def __init__(self, tp=None):
        self.type = tp
        self.name = 'list'

    def __call__(self, tp):
        assert self.type is None
        return List(tp)


cdef class Tuple:
    def __init__(self, *tps):
        if len(tps) == 1 and not isinstance(tps[0], Iterable):
            self.type = tps[0]
        else:
            self.type = tps if tps else None
        self.name = 'tuple'

    def __call__(self, *tps):
        assert self.type is None
        return Tuple(*tps)


cdef class Dict:
    def __init__(self, key_type=None, value_type=None):
        self.key_type = key_type
        self.value_type = value_type
        self.name = 'dict'

    def __call__(self, key_type, value_type):
        assert self.key_type is None and self.value_type is None
        return Dict(key_type, value_type)


cdef class Reference:
    def __init__(self, model):
        self.model = model
        self.name = 'reference'


cdef class OneOf:
    def __init__(self, *references):
        self.references = references


cdef class ValueType:
    bool = PrimitiveType.bool
    int8 = PrimitiveType.int8
    int16 = PrimitiveType.int16
    int32 = PrimitiveType.int32
    int64 = PrimitiveType.int64
    uint8 = PrimitiveType.uint8
    uint16 = PrimitiveType.uint16
    uint32 = PrimitiveType.uint32
    uint64 = PrimitiveType.uint64
    float16 = PrimitiveType.float16
    float32 = PrimitiveType.float32
    float64 = PrimitiveType.float64
    bytes = PrimitiveType.bytes
    unicode = PrimitiveType.unicode
    string = PrimitiveType.unicode
    complex64 = PrimitiveType.complex64
    complex128 = PrimitiveType.complex128
    slice = ExtendType.slice
    arr = ExtendType.arr
    dtype = ExtendType.dtype
    key = ExtendType.key
    datetime64 = ExtendType.datetime64
    timedelta64 = ExtendType.timedelta64
    index = ExtendType.index
    series = ExtendType.series
    dataframe = ExtendType.dataframe
    function = ExtendType.function
    tzinfo = ExtendType.tzinfo
    interval_arr = ExtendType.interval_arr
    freq = ExtendType.freq
    namedtuple = ExtendType.namedtuple
    regex = ExtendType.regex
    pickled = ExtendType.pickled

    identity = Identity()

    list = List()
    tuple = Tuple()
    dict = Dict()

    reference = Reference
    oneof = OneOf


cdef class Field:
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        self._model_cls = None
        self.attr = None

        self.tag = tag
        self._tag_name = None
        if not callable(self.tag):
            self._tag_name = self.tag
        self.default_val = default
        self.weak_ref = weak_ref
        self.on_serialize = on_serialize
        self.on_deserialize = on_deserialize

    @property
    def model(self):
        return self._model_cls

    @model.setter
    def model(self, model_cls):
        self._model_cls = model_cls

    cpdef str tag_name(self, Provider provider):
        if self._tag_name is None:
            return self.tag(provider)
        return self._tag_name

    cpdef serialize(self, Provider provider, model_instance, obj):
        return provider.serialize_field(self, model_instance, obj)

    cpdef deserialize(self, Provider provider, model_instance, obj, list callbacks, dict key_to_instance):
        return provider.deserialize_field(self, model_instance, obj, callbacks, key_to_instance)

    @property
    def type(self):
        return self._type

    @property
    def default(self):
        return self.default_val


cdef class AnyField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = None


cdef class IdentityField(Field):
    def __init__(self, tag, tp=None, default=None, bint weak_ref=False,
                 on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        if tp is not None:
            self._type = ValueType.identity(tp)
        else:
            self._type = ValueType.identity


cdef class BoolField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.bool


cdef class Int8Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.int8


cdef class Int16Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.int16


cdef class Int32Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.int32


cdef class Int64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.int64


cdef class UInt8Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.uint8


cdef class UInt16Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.uint16


cdef class UInt32Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.uint32


cdef class UInt64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.uint64


cdef class Float16Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.float64


cdef class Float32Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.float32


cdef class Float64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.float64


cdef class Complex64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.complex64


cdef class Complex128Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.complex128


cdef class StringField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.unicode


cdef class BytesField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.bytes


cdef class UnicodeField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.unicode


cdef class KeyField(Field):
    # this field is to store the HasKey object
    # we only store the key when pickling the serializable object

    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.key


cdef class NDArrayField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.arr


cdef class Datetime64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.datetime64


cdef class Timedelta64Field(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.timedelta64


cdef class DataTypeField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.dtype


cdef class IndexField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.index


cdef class SeriesField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.series


cdef class DataFrameField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.dataframe


cdef class SliceField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.slice


cdef class FunctionField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.function


cdef class NamedtupleField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.namedtuple


cdef class TZInfoField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.tzinfo


cdef class IntervalArrayField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.interval_arr


cdef class RegexField(Field):
    def __init__(self, tag, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.regex


cdef inline _handle_nest_reference(field, ref):
    if not isinstance(ref, Reference):
        return ref

    reference_field = ReferenceField(None, ref.model)
    reference_field.model = field.model
    return reference_field.type


cdef class ListField(Field):
    def __init__(self, tag, tp=None, default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        if tp is not None and type(tp) == Reference:
            self._type = None
            self._nest_ref = tp
        else:
            self._type = ValueType.list(tp)

    @property
    def model(self):
        return self._model_cls

    @model.setter
    def model(self, new_model_cls):
        if getattr(self, '_nest_ref', None) is not None and \
                self._nest_ref.model == 'self' and self._model_cls is not None and \
                new_model_cls is not None:
            raise SelfReferenceOverwritten('self reference is overwritten')
        self._model_cls = new_model_cls

    @property
    def type(self):
        if self._type is None:
            self._type = ValueType.list(
                _handle_nest_reference(self, self._nest_ref))
        return self._type


cdef class TupleField(Field):
    def __init__(self, tag, *tps, **kwargs):
        default = kwargs.pop('default', None)
        weak_ref = kwargs.pop('weak_ref', False)
        on_serialize = kwargs.pop('on_serialize', None)
        on_deserialize = kwargs.pop('on_deserialize', None)
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        if len(tps) == 1 and isinstance(tps[0], Iterable):
            tps = tps[0]
        self._type = ValueType.tuple(*tps)


cdef class DictField(Field):
    def __init__(self, tag, key_type=None, value_type=None,
                 default=None, bint weak_ref=False, on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        self._type = ValueType.dict(key_type, value_type)


cdef class ReferenceField(Field):
    def __init__(self, tag, model, default=None, bint weak_ref=False,
                 on_serialize=None, on_deserialize=None):
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)
        if model is None:
            self._type = ValueType.reference(None)
            self._model = None
        elif not isinstance(model, str):
            self._type = ValueType.reference(model)
        else:
            self._type = None
            self._model = model

    @property
    def model(self):
        return self._model_cls

    @model.setter
    def model(self, new_model_cls):
        if getattr(self, '_model', None) == 'self' and \
                self._model_cls is not None and new_model_cls is not None:
            raise SelfReferenceOverwritten('self reference is overwritten')
        self._model_cls = new_model_cls

    @property
    def type(self):
        if not self._type:
            if self._model == 'self':
                self._type = ValueType.reference(self.model)
            elif '.' not in self._model:
                model = getattr(inspect.getmodule(self.model), self._model)
                self._type = ValueType.reference(model)
            else:
                module, name = self._model.rsplit('.', 1)
                model = getattr(importlib.import_module(module), name)
                self._type = ValueType.reference(model)

        return self._type


cdef class OneOfField(Field):
    def __init__(self, tag, **kw):
        default = kw.pop('default', None)
        weak_ref = kw.pop('weak_ref', False)
        on_serialize = kw.pop('on_serialize', None)
        on_deserialize = kw.pop('on_deserialize', None)
        super().__init__(
            tag, default=default, weak_ref=weak_ref,
            on_serialize=on_serialize, on_deserialize=on_deserialize)

        self.fields = [ReferenceField(tag, model)
                       for tag, model in kw.items()]

    @property
    def type(self):
        if self._type is None:
            self._type = ValueType.oneof(*[f.type for f in self.fields])
        return self._type

    @property
    def attrs(self):
        return [f.attr for f in self.fields]


cdef inline set_model(dict fields, cls):
    cdef str slot
    cdef bint modified

    for slot, field in fields.items():
        if not isinstance(field, OneOfField):
            try:
                field.model = cls
            except SelfReferenceOverwritten:
                field = copy.copy(field)
                # reset old model after copy
                field.model = None
                field.model = cls
                cls._FIELDS[slot] = field
        else:
            one_field_fields = []
            modified = False
            for f in field.fields:
                try:
                    f.model = cls
                except SelfReferenceOverwritten:
                    f = copy.copy(f)
                    # reset old model after copy
                    f.model = None
                    f.model = cls
                    modified = True
                f.attr = field.attr
                one_field_fields.append(f)
            if modified:
                field.fields = one_field_fields


class SerializableMetaclass(type):
    def __new__(mcs, str name, tuple bases, dict kv):
        cdef list slots
        cdef set sslots
        cdef object props
        cdef dict fields
        cdef str k, ser_name
        cdef object v, ser_index, mod_name

        slots = list(kv.pop('__slots__', ()))
        sslots = set(slots)

        props = dict()
        for base in bases:
            if hasattr(base, '_FIELDS'):
                props.update(base._FIELDS)
        props.update(kv)
        props = OrderedDict(sorted(props.items(),
                                   key=lambda x: x[0] in kv, reverse=True))

        fields = {}
        for k, v in props.items():
            if not isinstance(v, Field):
                continue

            if k not in sslots:
                slots.append(k)

            if isinstance(v, IdentityField):
                if '_ID_FIELD' not in kv:
                    kv['_ID_FIELD'] = (v,)
                else:
                    kv['_ID_FIELD'] += (v,)

            fields[k] = v
            v.attr = k

            if k in kv:
                del kv[k]

        kv['_FIELDS'] = fields
        kv['__slots__'] = tuple(slots)

        cls = type.__new__(mcs, name, bases, kv)
        set_model(fields, cls)
        try:
            ser_index = kv['__serializable_index__']
        except KeyError:
            ser_name = name
            mod_name = kv.get('__module__', None)
            if mod_name:
                ser_name = mod_name.split('.', 1)[0] + '.' + ser_name
            ser_index = cls.__serializable_index__ = mmh_hash(to_binary(ser_name), signed=False)
        _serializable_registry[ser_index] = cls
        return cls


class Serializable(metaclass=SerializableMetaclass):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        for slot, arg in zip(self.__slots__, args):
            object.__setattr__(self, slot, arg)

        for key, val in kwargs.items():
            object.__setattr__(self, key, val)

    @classmethod
    def cls(cls, Provider provider):
        if provider.type == ProviderType.json:
            return dict

        provider_name = ProviderType(provider.type).name
        raise TypeError(f'Unknown provider type `{provider_name}` for class `{cls.__name__}`')

    def serialize(self, Provider provider, obj=None):
        return provider.serialize_model(self, obj=obj)

    @classmethod
    def deserialize(cls, Provider provider, obj, list callbacks=None, dict key_to_instance=None):
        call_cb = callbacks is None
        callbacks = [] if callbacks is None else callbacks
        key_to_instance = {} if key_to_instance is None else key_to_instance
        obj = provider.deserialize_model(cls, obj, callbacks, key_to_instance)
        if call_cb:
            [cb(key_to_instance) for cb in callbacks]
        return obj

    def to_pb(self, obj=None, data_serial_type=None, pickle_protocol=None):
        from .pbserializer import ProtobufSerializeProvider

        provider = ProtobufSerializeProvider(data_serial_type=data_serial_type,
                                             pickle_protocol=pickle_protocol)
        return self.serialize(provider, obj=obj)

    @classmethod
    def from_pb(cls, obj):
        from .pbserializer import ProtobufSerializeProvider

        provider = ProtobufSerializeProvider()
        return cls.deserialize(provider, obj)

    def to_json(self, obj=None, data_serial_type=None, pickle_protocol=None):
        from .jsonserializer import JsonSerializeProvider

        provider = JsonSerializeProvider(data_serial_type=data_serial_type,
                                         pickle_protocol=pickle_protocol)
        return self.serialize(provider, obj=obj)

    @classmethod
    def from_json(cls, obj):
        from .jsonserializer import JsonSerializeProvider

        provider = JsonSerializeProvider()
        return cls.deserialize(provider, obj)


cdef class AttrWrapper:
    def __init__(self, obj):
        self._obj = obj

    cpdef asdict(self):
        return dict(self._obj)

    def __getattr__(self, item):
        return self._obj[item]

    def __setattr__(self, key, value):
        self._obj[key] = value


class AttributeAsDict(Serializable):
    attr_tag = None

    @classmethod
    def cls(cls, Provider provider):
        if provider.type == ProviderType.protobuf and cls.attr_tag is None:
            from .protos.value_pb2 import Value
            return Value
        return super().cls(provider)

    def serialize(self, Provider provider, obj=None):
        return provider.serialize_attribute_as_dict(self, obj=obj)

    @classmethod
    def deserialize(cls, Provider provider, obj, list callbacks=None, dict key_to_instance=None):
        call_cb = callbacks is None
        callbacks = [] if callbacks is None else callbacks
        key_to_instance = {} if key_to_instance is None else key_to_instance
        obj = provider.deserialize_attribute_as_dict(
            cls, obj, callbacks, key_to_instance)
        if call_cb:
            [cb(key_to_instance) for cb in callbacks]
        return obj


class HasKey(object):
    __slots__ = '_key', '_id'


class HasData(object):
    __slots__ = '_data',


cdef class KeyPlaceholder:
    def __init__(self, key, id):
        self.key = key
        self.id = id


cpdef serializes(Provider provider, objects):
    return [obj.serialize(provider) if obj is not None else None for obj in objects]


cpdef list deserializes(Provider provider, list models, list objects):
    cdef list callbacks
    cdef dict key_to_instance
    cdef list objs

    callbacks = []
    key_to_instance = {}
    objs = []
    for model, pb_obj in zip(models, objects):
        if model is type(None):
            objs.append(None)
            continue
        objs.append(model.deserialize(provider, pb_obj, callbacks, key_to_instance))
    [cb(key_to_instance) for cb in callbacks]
    return objs


cpdef object get_serializable_by_index(object index):
    return _serializable_registry.get(index)


cpdef dict get_serializables():
    return _serializable_registry.copy()


cdef class Provider:
    cpdef serialize_field(self, Field field, model_instance, obj):
        raise NotImplementedError

    cpdef serialize_model(self, model_instance, obj=None):
        if obj is None:
            obj = model_instance.cls(self)()

        for name, field in model_instance._FIELDS.items():
            field.serialize(self, model_instance, obj)

        return obj

    cpdef serialize_attribute_as_dict(self, model_instance, obj=None):
        return self.serialize_model(model_instance, obj=obj)

    def deserialize_field(self, Field field, model_instance, obj, list callbacks, dict key_to_instance):
        raise NotImplementedError

    cpdef deserialize_model(self, model_cls, obj, list callbacks, dict key_to_instance):
        cdef object kw
        cdef IdentityField id_field
        cdef object model_instance
        cdef str name
        cdef Field field

        kw = AttrWrapper(dict())
        if hasattr(model_cls, '_ID_FIELD'):
            for id_field in model_cls._ID_FIELD:
                id_field.deserialize(self, kw, obj, callbacks, key_to_instance)
        model_instance = model_cls(**kw.asdict())

        for name, field in model_instance._FIELDS.items():
            field.deserialize(self, model_instance, obj, callbacks, key_to_instance)

        if hasattr(model_instance, 'key') and hasattr(model_instance, 'id'):
            key_to_instance[model_instance.key, model_instance.id] = model_instance
        return model_instance

    def deserialize_attribute_as_dict(self, model_cls, obj, list callbacks, dict key_to_instance):
        return self.deserialize_model(model_cls, obj, callbacks, key_to_instance)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import AnyField, BoolField
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, parse_index


class DataFrameSetIndex(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_SET_INDEX

    _keys = AnyField('keys')
    _drop = BoolField('drop')
    _append = BoolField('append')
    _verify_integrity = BoolField('verify_integrity')

    def __init__(self, keys=None, drop=True, append=False, verify_integrity=False,
                 output_types=None, **kw):
        super().__init__(_keys=keys, _drop=drop, _append=append,
                         _verify_integrity=verify_integrity, _output_types=output_types, **kw)

    @property
    def keys(self):
        return self._keys

    @property
    def drop(self):
        return self._drop

    @property
    def append(self):
        return self._append

    @property
    def verify_integrity(self):
        return self._verify_integrity

    def __call__(self, df):
        new_df = build_empty_df(df.dtypes).set_index(keys=self.keys, drop=self.drop, append=self.append,
                                                     verify_integrity=self.verify_integrity)
        return self.new_dataframe([df], shape=(df.shape[0], new_df.shape[1]), dtypes=new_df.dtypes,
                                  index_value=parse_index(new_df.index),
                                  columns_value=parse_index(new_df.columns, store_data=True))

    @classmethod
    def _tile_column_axis_n_chunk(cls, op, in_df, out_df, out_chunks):
        if not isinstance(op.keys, str):  # pragma: no cover
            raise NotImplementedError('DataFrame.set_index only support label')
        if op.verify_integrity:  # pragma: no cover
            raise NotImplementedError('DataFrame.set_index not support verify_integrity yet')

        try:
            column_index = in_df.columns_value.to_pandas().get_loc(op.keys)
        except KeyError:  # pragma: no cover
            raise NotImplementedError('The new index label must be a column of the original dataframe')

        chunk_index = np.searchsorted(np.cumsum(in_df.nsplits[1]), column_index + 1)

        for row_idx in range(in_df.chunk_shape[0]):
            index_chunk = in_df.cix[row_idx, chunk_index]
            for col_idx in range(in_df.chunk_shape[1]):
                input_chunk = in_df.cix[row_idx, col_idx]
                if op.drop and input_chunk.key == index_chunk.key:
                    new_shape = (input_chunk.shape[0], input_chunk.shape[1] - 1)
                    columns = parse_index(input_chunk.columns_value.to_pandas().drop(op.keys), store_data=True)
                else:
                    new_shape = input_chunk.shape
                    columns = input_chunk.columns_value
                out_op = op.copy().reset_key()
                out_chunk = out_op.new_chunk([index_chunk, input_chunk],
                                             shape=new_shape, dtypes=out_df.dtypes, index=input_chunk.index,
                                             index_value=parse_index(pd.Int64Index([])),
                                             columns_value=columns)
                out_chunks.append(out_chunk)

    @classmethod
    def _tile_column_axis_1_chunk(cls, op, in_df, out_df, out_chunks):
        out_pd_index = out_df.index_value.to_pandas()
        for c in in_df.chunks:
            chunk_op = op.copy().reset_key()
            chunk_shape = (c.shape[0], out_df.shape[1])
            index_value = parse_index(out_pd_index, c)
            out_chunk = chunk_op.new_chunk([c], shape=chunk_shape,
                                           dtypes=out_df.dtypes, index=c.index,
                                           index_value=index_value,
                                           columns_value=out_df.columns_value)
            out_chunks.append(out_chunk)

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        out_chunks = []
        if in_df.chunk_shape[1] > 1:
            cls._tile_column_axis_n_chunk(op, in_df, out_df, out_chunks)
        else:
            cls._tile_column_axis_1_chunk(op, in_df, out_df, out_chunks)

        new_op = op.copy()
        columns_nsplits = list(in_df.nsplits[1])
        if op.drop:
            columns_nsplits = tuple(split - 1 if i == 0 else split for i, split in enumerate(columns_nsplits))
        nsplits = (in_df.nsplits[0], columns_nsplits)
        return new_op.new_dataframes(op.inputs, out_df.shape, dtypes=out_df.dtypes,
                                     index_value=out_df.index_value,
                                     columns_value=out_df.columns_value,
                                     chunks=out_chunks, nsplits=nsplits)

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]

        if len(op.inputs) == 2:
            # axis 1 has more than 1 chunk
            index_chunk, input_chunk = op.inputs
            # Optimization: we don't need to get value of the column
            # that is set as new index.
            if input_chunk.key == index_chunk.key:
                new_index = op.keys
            else:
                new_index = ctx[index_chunk.key][op.keys]
            ctx[chunk.key] = ctx[input_chunk.key].set_index(
                new_index, drop=op.drop, append=op.append,
                verify_integrity=op.verify_integrity)
        else:
            # axis 1 has 1 chunk
            inp = ctx[op.inputs[0].key]
            ctx[chunk.key] = inp.set_index(
                op.keys, drop=op.drop, append=op.append,
                verify_integrity=op.verify_integrity)


def set_index(df, keys, drop=True, append=False, inplace=False, verify_integrity=False):
    op = DataFrameSetIndex(keys=keys, drop=drop, append=append,
                           verify_integrity=verify_integrity, output_types=[OutputType.dataframe])
    result = op(df)
    if not inplace:
        return result
    else:
        df.data = result.data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import itertools
import operator
import uuid
from collections import OrderedDict
from collections.abc import Iterable
from functools import lru_cache, reduce, wraps
from math import ceil
from numbers import Integral

import numpy as np
try:
    import tiledb
except (ImportError, OSError):  # pragma: no cover
    tildb = None

from ..core import ExecutableTuple
from ..utils import lazy_import
from ..lib.mmh3 import hash_from_buffer

cp = lazy_import('cupy', globals=globals(), rename='cp')


def normalize_shape(shape):
    if isinstance(shape, Iterable):
        return tuple(shape)
    else:
        return shape,


def normalize_chunk_sizes(shape, chunk_size):
    shape = normalize_shape(shape)
    if not isinstance(chunk_size, tuple):
        if isinstance(chunk_size, Iterable):
            chunk_size = tuple(chunk_size)
        elif isinstance(chunk_size, int):
            chunk_size = (chunk_size,) * len(shape)

    if len(shape) != len(chunk_size):
        raise ValueError('Chunks must have the same dimemsion, '
                         f'got shape: {shape}, chunks: {chunk_size}')

    chunk_sizes = []
    for size, chunk in zip(shape, chunk_size):
        if isinstance(chunk, Iterable):
            if not isinstance(chunk, tuple):
                chunk = tuple(chunk)

            if sum(chunk) != size:
                raise ValueError('chunks shape should be of the same length, '
                                 f'got shape: {size}, chunks: {chunk}')
            chunk_sizes.append(chunk)
        else:
            assert isinstance(chunk, int)

            if size == 0:
                sizes = (0,)
            else:
                sizes = tuple(chunk for _ in range(int(size / chunk))) + \
                    (tuple() if size % chunk == 0 else (size % chunk,))
            chunk_sizes.append(sizes)

    return tuple(chunk_sizes)


def broadcast_shape(*shapes):
    if len(shapes) == 1:
        return shapes[0]

    out_shapes = []
    for ss in itertools.zip_longest(*[reversed(s) for s in shapes], fillvalue=-1):
        shape = max(s for s in ss if s != -1)
        if any(i != -1 and i != 1 and i != shape and not np.isnan(i) for i in ss):
            raise ValueError('Operands could not be broadcast together '
                             'with shape {0}'.format(' '.join(map(str, shapes))))
        out_shapes.append(shape)
    return tuple(reversed(out_shapes))


def get_chunk_slices(nsplits, idx):
    return tuple(slice(sum(nsplit[:idx]), sum(nsplit[:idx + 1]))
                 for idx, nsplit in zip(idx, nsplits))


def gen_random_seeds(n, random_state):
    assert isinstance(random_state, np.random.RandomState)
    return tuple(np.frombuffer(random_state.bytes(n * 4), dtype=np.uint32).tolist())


def validate_axis(ndim, axis, argname=None):
    if axis >= ndim or axis < -ndim:
        raise np.AxisError(axis, ndim=ndim, msg_prefix=argname)

    return axis if axis >= 0 else ndim + axis


def normalize_axis_tuple(axis, ndim, argname=None, allow_duplicate=False):
    """
    Normalizes an axis argument into a tuple of non-negative integer axes.

    This handles shorthands such as ``1`` and converts them to ``(1,)``,
    as well as performing the handling of negative indices covered by
    `normalize_axis_index`.

    By default, this forbids axes from being specified multiple times.

    Used internally by multi-axis-checking logic.

    Parameters
    ----------
    axis : int, iterable of int
        The un-normalized index or indices of the axis.
    ndim : int
        The number of dimensions of the array that `axis` should be normalized
        against.
    argname : str, optional
        A prefix to put before the error message, typically the name of the
        argument.
    allow_duplicate : bool, optional
        If False, the default, disallow an axis from being specified twice.

    Returns
    -------
    normalized_axes : tuple of int
        The normalized axis index, such that `0 <= normalized_axis < ndim`

    Raises
    ------
    AxisError
        If any axis provided is out of range
    ValueError
        If an axis is repeated

    See also
    --------
    normalize_axis_index : normalizing a single scalar axis
    """
    # Optimization to speed-up the most common cases.
    if type(axis) not in (tuple, list):
        try:
            axis = [operator.index(axis)]
        except TypeError:
            pass
    # Going via an iterator directly is slower than via list comprehension.
    axis = tuple([validate_axis(ndim, ax, argname) for ax in axis])
    if not allow_duplicate and len(set(axis)) != len(axis):
        if argname:
            raise ValueError(f'repeated axis in `{argname}` argument')
        else:
            raise ValueError('repeated axis')
    return axis


def validate_order(dtype, order):
    if getattr(dtype, 'fields', None) is None:
        if order is not None:
            raise ValueError('Cannot specify order when the array has no fields')
        else:
            return

    need_check = True
    if order is None:
        order = list(dtype.names)
        need_check = False
    elif isinstance(order, (list, tuple)):
        order = list(order)
    else:
        order = [order]
    if need_check:
        for o in order:
            if o not in dtype.fields:
                raise ValueError(f'unknown field name: {o}')
    return order


def inject_dtype(dtype):
    def inner(func):
        @wraps(func)
        def call(*tensors, **kw):
            kw['dtype'] = np.dtype(dtype)
            ret = func(*tensors, **kw)
            if ret is NotImplemented:
                reverse_func = getattr(inspect.getmodule(func), f'r{func.__name__}', None)
                if reverse_func is not None:
                    ret = reverse_func(*tensors[::-1], **kw)
                if ret is NotImplemented:
                    raise TypeError(
                        "unsupported operand type(s) for {0}: '{1}' and '{2}".format(
                            func.__name__, *[type(t) for t in tensors]))
            return ret

        return call

    return inner


def infer_dtype(np_func, empty=True, reverse=False, check=True):
    def make_arg(arg):
        if empty:
            return np.empty((1,) * max(1, arg.ndim), dtype=arg.dtype)
        else:
            if hasattr(arg, 'op') and hasattr(arg.op, 'data'):
                arg = arg.op.data
            return arg[(0,) * max(1, arg.ndim)]

    tensor_ufunc = '__tensor_ufunc__'

    def is_arg(arg):
        if hasattr(arg, tensor_ufunc):
            return False
        return hasattr(arg, 'ndim') and hasattr(arg, 'dtype')

    def inner(func):
        @wraps(func)
        def h(*tensors, **kw):
            usr_dtype = np.dtype(kw.pop('dtype')) if 'dtype' in kw else None
            args = [make_arg(t) if is_arg(t) else t for t in tensors]
            if reverse:
                args = args[::-1]
            np_kw = dict((k, make_arg(v) if hasattr(v, 'op') else v) for k, v in kw.items()
                         if is_arg(v) and k != 'out')

            dtype = None
            if not any(hasattr(arg, tensor_ufunc)
                       for arg in itertools.chain(args, np_kw.values())):
                # skip infer if encounter mars DataFrame etc
                # that implements __tensor_ufunc__
                try:
                    with np.errstate(all='ignore'):
                        dtype = np_func(*args, **np_kw).dtype
                except:  # noqa: E722
                    dtype = None

            if usr_dtype and dtype:
                can_cast_kwargs = {}
                if kw.get('casting') is not None:
                    can_cast_kwargs['casting'] = kw.get('casting')
                if check and not np.can_cast(dtype, usr_dtype, **can_cast_kwargs):
                    raise TypeError('No loop matching the specified signature '
                                    f'and casting was found for ufunc {np_func}')
                kw['dtype'] = usr_dtype
            else:
                kw['dtype'] = dtype

            ret = func(*tensors, **kw)
            if ret is NotImplemented:
                reverse_func = getattr(inspect.getmodule(func), f'r{func.__name__}', None) \
                    if not reverse else None
                if reverse_func is not None:
                    ret = reverse_func(*tensors[::-1], **kw)
                if ret is NotImplemented:
                    raise TypeError(
                        "unsupported operand type(s) for {0}: '{1}' and '{2}".format(
                            func.__name__, *[type(t) for t in tensors]))
            return ret

        return h

    return inner


def index_ndim(index):
    from .core import Tensor

    if isinstance(index, Tensor) and index.dtype == np.bool_:
        # boolean indexing will occupy the ndim
        return index.ndim

    return 1 if index is not None else 0


def replace_ellipsis(index, ndim):
    all_illipsis = list(i for i, idx in enumerate(index) if idx is Ellipsis)
    if len(all_illipsis) > 1:
        raise IndexError("an index can only have a single ellipsis ('...')")
    if not all_illipsis:
        return index

    illipsis_index = all_illipsis[0]
    n_extra = ndim - sum([index_ndim(i) for i in index]) + 1
    return index[:illipsis_index] + (slice(None),) * n_extra + index[illipsis_index + 1:]


def calc_sliced_size(size, sliceobj):
    if np.isnan(size):
        return np.nan

    start, stop, step = sliceobj.indices(size)
    return int(ceil(abs((stop - start) / float(step))))


def calc_object_length(obj, size=None):
    if np.isscalar(obj):
        return 1
    elif isinstance(obj, slice):
        return calc_sliced_size(size, obj)
    else:
        return len(obj)


def slice_split(index, sizes):
    size = sum(sizes)

    if isinstance(index, Integral):
        index = index if index >= 0 else size + index
        i = 0
        ind = index
        lens = list(sizes)
        while ind >= lens[0]:
            i += 1
            ind -= lens.pop(0)
        return {i: ind}

    assert isinstance(index, slice)
    start, stop, step = index.indices(size)

    slice_all = slice(None)

    if index == slice_all:
        return dict((k, slice_all) for k in range(len(sizes)))

    d = dict()
    if step > 0:
        for i, length in enumerate(sizes):
            if start < length and stop > 0:
                d[i] = slice(start, min(stop, length), step)
                start = (start - length) % step
            else:
                start = start - length
            stop -= length
    else:
        rstart = start  # running start
        chunk_boundaries = np.cumsum(sizes)
        for i, chunk_stop in reversed(list(enumerate(chunk_boundaries))):
            # create a chunk start and stop
            if i == 0:
                chunk_start = 0
            else:
                chunk_start = chunk_boundaries[i - 1]

            # if our slice is in this chunk
            if (chunk_start <= rstart < chunk_stop) and (rstart > stop):
                d[i] = slice(rstart - chunk_stop,
                             max(chunk_start - chunk_stop - 1,
                                 stop - chunk_stop),
                             step)

                # compute the next running start point,
                offset = (rstart - (chunk_start - 1)) % step
                rstart = chunk_start + offset - 1

    # replace 0:20:1 with : if appropriate
    for k, v in d.items():
        if v == slice(0, sizes[k], 1):
            d[k] = slice(None, None, None)

    if not d:  # special case x[:0]
        d[0] = slice(0, 0, 1)

    return d


def is_asc_sorted(arr):
    arr = np.asarray(arr)
    if len(arr) == 0:
        return True
    return np.all(arr[:-1] <= arr[1:])


def split_indexes_into_chunks(nsplits, indexes, ret_is_asc=True):
    indexes = np.asarray(indexes)
    chunk_idxes = np.empty_like(indexes)
    cum_nsplits = [np.cumsum(nsplit) for nsplit in nsplits]
    for i, cum_nsplit, index in zip(itertools.count(0), cum_nsplits, indexes):
        # handle negative value in index
        if hasattr(index, 'flags') and not index.flags.writeable:
            index = index.copy()
        index = np.add(index, cum_nsplit[-1], out=index, where=index < 0)
        sorted_idx = np.argsort(index)

        if np.any(index >= cum_nsplit[-1]):
            idx = index[index >= cum_nsplit[-1]][0]
            raise IndexError(f'index {idx} is out of bounds with size {cum_nsplit[-1]}')

        chunk_idx = np.searchsorted(cum_nsplit, index[sorted_idx], side='right')
        chunk_idxes[i, sorted_idx] = chunk_idx

    chunk_idxes_asc = False
    if ret_is_asc:
        chunk_idxes_asc = is_asc_sorted(np.lexsort(chunk_idxes[::-1]))

    chunk_index_to_indexes = OrderedDict()
    chunk_index_to_poses = OrderedDict()
    poses = np.arange(len(indexes[0]))
    for idx in itertools.product(*(range(len(nsplit)) for nsplit in nsplits)):
        cond = (chunk_idxes == np.array(idx).reshape((len(idx), 1))).all(axis=0)
        filtered = indexes[:, cond]
        for i in range(len(indexes)):
            filtered[i] = filtered[i] - (cum_nsplits[i][idx[i] - 1] if idx[i] > 0 else 0)
        chunk_index_to_indexes[idx] = filtered
        chunk_index_to_poses[idx] = poses[cond]

    if ret_is_asc:
        return chunk_index_to_indexes, chunk_index_to_poses, chunk_idxes_asc
    return chunk_index_to_indexes, chunk_index_to_poses


def calc_pos(fancy_index_shape, pos, xp=np):
    if isinstance(pos, dict):
        pos = xp.concatenate(list(pos.values()))
    select_pos = xp.empty(fancy_index_shape, dtype=int)
    select_pos.flat[pos] = xp.arange(select_pos.size)
    return select_pos


def decide_unify_split(*splits):
    # TODO (jisheng): In the future, we need more sophisticated way to decide the rechunk split
    # right now, for (2, 2) and (3, 1), we get the rechunk split as (2, 1, 1)
    if not splits:
        return ()
    raw_splits = splits
    # support broadcasting rules
    # decide_unify_splits((1,), (5,))  --> (5,)
    splits = set(s for s in splits if ((len(s) > 1) or (len(s) == 1 and s[0] != 1)))
    if len(splits) == 1:
        return splits.pop()
    if len(splits) == 0:
        return raw_splits[0]

    if any(np.isnan(sum(s)) for s in splits):
        raise ValueError(f'Tensor chunk sizes are unknown: {splits}')
    if len(set(sum(s) for s in splits)) > 1:
        raise ValueError(f'Splits do not hava same size: {splits}')

    q = [list(s) for s in splits]
    size = sum(q[0])
    cum = 0

    res = []
    while cum < size:
        m = min(s[0] for s in q)
        res.append(m)
        for s in q:
            s[0] -= m
            if s[0] == 0:
                s.pop(0)

        cum += m

    return tuple(res)


def unify_nsplits(*tensor_axes):
    tensor_splits = [dict((a, split) for a, split in zip(axes, t.nsplits) if split != (1,))
                     for t, axes in tensor_axes if t.nsplits]
    common_axes = reduce(operator.and_, [set(ts.keys()) for ts in tensor_splits]) if tensor_splits else set()
    axes_unified_splits = dict((ax, decide_unify_split(*(t[ax] for t in tensor_splits)))
                               for ax in common_axes)

    if len(common_axes) == 0:
        return tuple(t[0] for t in tensor_axes)

    res = []
    for t, axes in tensor_axes:
        new_chunk = dict((i, axes_unified_splits[ax]) for ax, i in zip(axes, range(t.ndim))
                         if ax in axes_unified_splits)
        res.append(t.rechunk(new_chunk)._inplace_tile())

    return tuple(res)


def unify_chunks(*tensors):
    tensor_axes = [(t, range(t.ndim)) if not isinstance(t, tuple) else t
                   for t in tensors]

    if len(tensor_axes) < 2:
        return tuple(t[0] if isinstance(t, tuple) else t for t in tensors)

    return unify_nsplits(*tensor_axes)


def check_out_param(out, t, casting):
    from .base import broadcast_to

    if not hasattr(out, 'shape'):
        raise TypeError('return arrays must be a tensor')

    try:
        broadcast_to(t, out.shape)
    except ValueError:
        raise ValueError("operands could not be broadcast together "
                         "with shapes ({0}) ({1})".format(','.join(str(s) for s in t.shape),
                                                          ','.join(str(s) for s in out.shape)))

    if not np.can_cast(t.dtype, out.dtype, casting):
        raise TypeError(f"output (typecode '{t.dtype.char}') could not be coerced "
                        f"to provided output paramter (typecode '{out.dtype.char}') "
                        f"according to the casting rule ''{casting}''")


def dictify_chunk_size(shape, chunk_size):
    """
    Given chunk_size which may be a tuple or dict, return a dict type all the same.

    :param shape: tensor's shape
    :param chunk_size: if dict provided, it's dimension id to chunk size;
                       if provided, it's the chunk size for each dimension.
    :return: dict form of chunk_size
    """
    if chunk_size is not None:
        if isinstance(chunk_size, Iterable):
            if not isinstance(chunk_size, dict):
                chunk_size = {i: c for i, c in enumerate(chunk_size)}
        elif isinstance(chunk_size, int):
            chunk_size = {i: chunk_size for i in range(len(shape))}
        else:
            raise TypeError(f'chunks must be iterable, got {type(chunk_size)}')

    if chunk_size is None:
        chunk_size = dict()

    return chunk_size


def decide_chunk_sizes(shape, chunk_size, itemsize):
    """
    Decide how a given tensor can be split into chunk.

    :param shape: tensor's shape
    :param chunk_size: if dict provided, it's dimension id to chunk size;
                       if provided, it's the chunk size for each dimension.
    :param itemsize: element size
    :return: the calculated chunk size for each dimension
    :rtype: tuple
    """

    from ..config import options

    chunk_size = dictify_chunk_size(shape, chunk_size)
    nleft = len(shape) - len(chunk_size)
    if nleft < 0:
        raise ValueError("chunks have more dimensions than input tensor")
    if nleft == 0:
        return normalize_chunk_sizes(shape, tuple(chunk_size[j] for j in range(len(shape))))

    max_chunk_size = options.chunk_store_limit

    # normalize the dimension which specified first
    dim_to_normalized = {i: normalize_chunk_sizes((shape[i],), (c,))[0]
                         for i, c in chunk_size.items()}

    left = {j: [] for j in range(len(shape)) if j not in dim_to_normalized}
    left_unsplit = {j: shape[j] for j in left}
    while True:
        nbytes_occupied = np.prod([max(c) for c in dim_to_normalized.values()]) * itemsize
        dim_size = np.maximum(int(np.power(max_chunk_size / nbytes_occupied, 1 / float(len(left)))), 1)
        for j, ns in left.copy().items():
            unsplit = left_unsplit[j]
            ns.append(int(np.minimum(unsplit, dim_size)))
            left_unsplit[j] -= ns[-1]
            if left_unsplit[j] <= 0:
                dim_to_normalized[j] = tuple(ns)
                del left[j]

        if len(left) == 0:
            break

    return tuple(dim_to_normalized[i] for i in range(len(dim_to_normalized)))


def check_random_state(seed):
    """
    Turn seed into a mt.random.RandomState instance

    :param seed:
        If seed is None, return the RandomState singleton used by mt.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.
    :return:
    """
    from . import random as mtrand
    from numpy import random as np_mtrand

    if seed is None or seed is mtrand or seed is np_mtrand:
        return mtrand._random_state
    if isinstance(seed, (Integral, np.integer)):
        return mtrand.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return mtrand.RandomState.from_numpy(seed)
    if isinstance(seed, mtrand.RandomState):
        return seed
    raise ValueError(f'{seed} cannot be used to seed a mt.random.RandomState'
                     ' instance')


def create_fetch_tensor(chunk_size, shape, dtype, tensor_key=None, tensor_id=None, chunk_keys=None):
    """
    Construct Fetch tensor on the fly, using given chunk_size, shape, dtype,
    as well as possible tensor_key, tensor_id and chunk keys.
    """
    from ..config import options
    from .fetch import TensorFetch

    if not isinstance(dtype, np.dtype):
        dtype = np.dtype(dtype)
    if chunk_keys is None:
        chunk_keys = itertools.repeat(None)

    # compute chunks
    chunk_size = chunk_size or options.chunk_size
    chunk_size = decide_chunk_sizes(shape, chunk_size, dtype.itemsize)
    chunk_size_idxes = (range(len(size)) for size in chunk_size)

    fetch_op = TensorFetch(dtype=dtype).reset_key()

    chunks = []
    for chunk_shape, chunk_idx, chunk_key in zip(itertools.product(*chunk_size),
                                                 itertools.product(*chunk_size_idxes),
                                                 chunk_keys):
        chunk = fetch_op.copy().reset_key().new_chunk(None, shape=chunk_shape, index=chunk_idx,
                                                      _key=chunk_key, hex=uuid.uuid4().hex)
        chunks.append(chunk)
    return fetch_op.copy().new_tensor(None, shape=shape, dtype=dtype, nsplits=chunk_size,
                                      chunks=chunks, _key=tensor_key, _id=tensor_id, hex=uuid.uuid4().hex)


def create_mutable_tensor(name, chunk_size, shape, dtype, chunk_keys=None, chunk_eps=None):
    """
    Construct MutableTensor on the fly, using given name, chunk_size, shape, dtype,
    as well as possible chunk keys.
    """
    from .core import MutableTensor, MutableTensorData
    tensor = create_fetch_tensor(chunk_size, shape, dtype, chunk_keys=chunk_keys)
    return MutableTensor(data=MutableTensorData(name=name, op=None, shape=shape, dtype=tensor.dtype,
                                                nsplits=tensor.nsplits, key=tensor.key, chunks=tensor.chunks,
                                                chunk_eps=chunk_eps))


def setitem_as_records(nsplits_acc, output_chunk, value, ts, is_scalar):
    """
    Turns a `__setitem__`  to a list of index-value records.

    Parameters:
    :arg nsplits_acc:
        Accumulate nsplits arrays of the output tensor chunks.

    :arg output_chunk:
        A chunk in the output of the `__setitem__` op.

    :arg value:
        The scalar or ndarray value that are set to the tensor.

    :arg ts:
        The timestamp value will be contained in the records.

    :arg is_scalar:
        Whether the value should be treat as scalar value, including tuple
        for structured arrays.

    :returns:
        A list of `[index, value, timestamp]`.
    """
    # prepare chunk value
    if is_scalar:
        chunk_value = value
    else:
        chunk_value_slice = tuple(slice(nsplits_acc[i][output_chunk.index[i]],
                                        nsplits_acc[i][output_chunk.index[i] + 1])
                                  for i in range(len(output_chunk.index)))
        chunk_value = value[chunk_value_slice]

    input_chunk = output_chunk.op.input

    input_indices = []  # index in the chunk of the mutable tensor
    value_indices = []  # index in the chunk of the assigned value
    for d, s in zip(output_chunk.op.indexes, input_chunk.shape):
        # expand the index (slice)
        idx = np.r_[slice(*d.indices(s)) if isinstance(d, slice) else d]
        input_indices.append(idx)
        if not isinstance(d, Integral):
            value_indices.append(np.arange(len(idx)))

    records = []
    for chunk_idx, value_idx in zip(itertools.product(*input_indices),
                                    itertools.product(*value_indices)):
        new_value = chunk_value if is_scalar else chunk_value[value_idx]
        records.append((np.ravel_multi_index(chunk_idx, input_chunk.shape), ts, new_value))
    return records


def filter_inputs(inputs):
    from ..core import Base, Entity

    return [inp for inp in inputs if isinstance(inp, (Base, Entity))]


# As TileDB Ctx's creation is a bit time-consuming,
# we just cache the Ctx
# also remember the arguments should be hashable
@lru_cache(10)
def _create_tiledb_ctx(conf_tuple):
    if conf_tuple is not None:
        return tiledb.Ctx(dict(conf_tuple))
    return tiledb.Ctx()


def get_tiledb_ctx(conf):
    key = tuple(conf.items()) if conf is not None else None
    return _create_tiledb_ctx(key)


# this function is only used for pandas' compatibility
def to_numpy(pdf):
    try:
        return pdf.to_numpy()
    except AttributeError:  # pragma: no cover
        return pdf.values


def check_order(order_str, available_options='KACF',
                err_msg='order not understood'):
    order_str = order_str.upper()
    if order_str not in available_options:
        raise TypeError(err_msg)


def get_order(order_str, to_keep_order, available_options='KACF',
              err_msg='order not understood'):
    from .core import TensorOrder

    check_order(order_str, available_options=available_options,
                err_msg=err_msg)

    if order_str in 'KA':
        return to_keep_order
    elif order_str == 'C':
        return TensorOrder.C_ORDER
    else:
        return TensorOrder.F_ORDER


def reverse_order(old_order):
    from .core import TensorOrder

    assert isinstance(old_order, TensorOrder)
    return TensorOrder.C_ORDER if old_order == TensorOrder.F_ORDER else TensorOrder.F_ORDER


def hash_on_axis(ar, axis, n_dest):
    ar = np.asarray(ar)
    # cannot be scalar
    assert ar.ndim > 0
    axis = validate_axis(ar.ndim, axis)

    if n_dest == 1:
        return np.zeros(ar.shape[axis], dtype=np.uint32)

    if ar.ndim > 2:
        ret = np.empty(ar.shape[axis], dtype=np.uint32)

        def _hash_to_dest(data):
            i = data[0]
            idx = (slice(None),) * axis + (i,)
            ret[i] = hash_from_buffer(memoryview(ar[idx])) % n_dest

        np.apply_along_axis(_hash_to_dest, 0, np.arange(ar.shape[axis])[np.newaxis, :])
        return ret
    else:
        def _hash_to_dest(data):
            return hash_from_buffer(memoryview(data)) % n_dest

        if ar.ndim == 1:
            ar = ar.reshape(ar.size, 1)
        return np.apply_along_axis(_hash_to_dest, 1 - axis, ar)


def fetch_corner_data(tensor, session=None):
    print_option = np.get_printoptions()
    # only fetch corner data when data > threshold
    threshold = print_option['threshold']
    # number of edge items to print
    edgeitems = print_option['edgeitems']

    # we fetch corner data based on the fact that
    # the tensor must have been executed,
    # thus the size could not be NaN
    if tensor.size > threshold:
        # two edges for each exis
        indices_iter = list(itertools.product(*(range(2) for _ in range(tensor.ndim))))
        corners = np.empty(shape=(2,) * tensor.ndim, dtype=object)
        shape = [0 for _ in range(tensor.ndim)]
        for indices in indices_iter:
            slc = []
            for ax, i in enumerate(indices):
                size = tensor.shape[ax]
                if size > edgeitems * 2 + 2:
                    # fetch two more elemens
                    if i == 0:
                        slc.append(slice(edgeitems + 1))
                    else:
                        slc.append(slice(-edgeitems - 1, None))
                    shape[ax] += edgeitems + 1
                else:
                    i_sep = size // 2
                    if i == 0:
                        slc.append(slice(i_sep))
                        shape[ax] += i_sep
                    else:
                        slc.append(slice(i_sep, None))
                        shape[ax] += size - i_sep
            corners[indices] = tensor[tuple(slc)]
        # fetch together
        fetched = ExecutableTuple(corners.flat).fetch(session=session)
        for indices, f in zip(indices_iter, fetched):
            corners[indices] = f
        return np.block(corners.tolist())
    else:
        return tensor.fetch(session=session)


def implement_scipy(scipy_fun):
    import re
    import textwrap

    def wrapper(fun):
        if scipy_fun is None:
            return None
        if not fun.__doc__:
            doc_str = textwrap.dedent(scipy_fun.__doc__)
            lines = []
            for line in doc_str.splitlines(keepends=False):
                # skip function headers
                if line.startswith(scipy_fun.__name__ + '('):
                    continue
                # skip version marks
                if line.strip().startswith('.. versionadded::'):
                    continue
                # skip examples
                if line.strip() == 'Examples':
                    break
                lines.append(line)
            doc_str = '\n'.join(lines).strip()
            # remove trailing empty sections
            fun.__doc__ = re.sub(r'[A-Za-z]+\n-+$', '', doc_str).strip()
        return fun

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from collections import namedtuple
from functools import lru_cache
from typing import Dict

import ray

from ..graph import DAG
from ..operands import Fetch
from ..tiles import get_tiled
from ..utils import build_fetch_chunk
from ..executor import Executor, GraphExecution


class _OperandWrapper:
    __slots__ = 'op', 'chunks'

    def __init__(self, op, chunks):
        """
        As we only serde op for Ray executors, but op only weakly reference chunks,
        So we create a wrapper here to keep the reference
        """
        self.op = op
        self.chunks = chunks


def operand_serializer(op):
    graph = DAG()
    inputs = [build_fetch_chunk(inp) for inp in op.inputs or []]
    new_op = op.copy()

    kws = []
    for c in op.outputs:
        params = c.params.copy()
        params['_key'] = c.key
        params.update(c.extra_params)
        kws.append(params)

    chunks = new_op.new_chunks(inputs, kws=kws, output_limit=len(kws))
    for obj in chunks + inputs:
        graph.add_node(obj)

    return graph.to_json()


def operand_deserializer(value):
    graph = DAG.from_json(value)
    if len(graph) == 1:
        chunks = [list(graph)[0]]
    else:
        chunks = [c for c in graph if not isinstance(c.op, Fetch)]
    op = chunks[0].op
    return _OperandWrapper(op, chunks)


@lru_cache(500)
def _register_ray_serializer(op):
    # register a custom serializer for Mars operand
    try:
        ray.register_custom_serializer(
            type(op), serializer=operand_serializer,
            deserializer=operand_deserializer)
    except AttributeError:  # ray >= 1.0
        try:
            from ray.worker import global_worker

            global_worker.check_connected()
            context = global_worker.get_serialization_context()
            context.register_custom_serializer(
                type(op), serializer=operand_serializer,
                deserializer=operand_deserializer)
        except AttributeError:  # ray >= 1.2.0
            ray.util.register_serializer(
                type(op), serializer=operand_serializer,
                deserializer=operand_deserializer)


class GraphExecutionForRay(GraphExecution):
    def handle_op(self, *args, **kw):
        return RayExecutor.handle(*args, **kw)


ChunkMeta = namedtuple('ChunkMeta', ['shape', 'object_id'])


@ray.remote
class RemoteMetaStore:
    def __init__(self):
        self._store = dict()

    def set_meta(self, chunk_key, meta):
        self._store[chunk_key] = meta

    def get_meta(self, chunk_key):
        return self._store[chunk_key]

    def get_shape(self, chunk_key):
        return self._store[chunk_key].shape

    def chunk_keys(self):
        return list(self._store.keys())

    def delete_keys(self, keys):
        if not isinstance(keys, list):
            keys = [keys]
        for k in keys:
            del self._store[k]


class RayStorage:
    """
    `RayStorage` is a dict-like class. When executed in local, Mars executor will store chunk result in a
    dict(chunk_key -> chunk_result), here uses Ray actor to store them as remote objects.
    """
    def __init__(self, meta_store=None):
        self.meta_store = meta_store or RemoteMetaStore.remote()

    def __getitem__(self, item):
        meta: ChunkMeta = ray.get(self.meta_store.get_meta.remote(item))
        return ray.get(meta.object_id)

    def __setitem__(self, key, value):
        object_id = ray.put(value)
        shape = getattr(value, 'shape', None)
        meta = ChunkMeta(shape=shape, object_id=object_id)
        set_meta = self.meta_store.set_meta.remote(key, meta)
        ray.wait([object_id, set_meta])

    def copy(self):
        return RayStorage(meta_store=self.meta_store)

    def update(self, mapping: Dict):
        tasks = []
        for k, v in mapping.items():
            object_id = ray.put(v)
            tasks.append(object_id)
            shape = getattr(v, 'shape', None)
            meta = ChunkMeta(shape=shape, object_id=object_id)
            set_meta = self.meta_store.set_meta.remote(k, meta)
            tasks.append(set_meta)
        ray.wait(tasks)

    def __iter__(self):
        return iter(ray.get(self.meta_store.chunk_keys.remote()))

    def __delitem__(self, key):
        ray.wait([self.meta_store.delete_keys.remote(key)])


@ray.remote
def execute_on_ray(func, results, op_wrapper: _OperandWrapper):
    op = op_wrapper.op
    func(results, op)


class RayExecutor(Executor):
    """
    Wraps the execute function as a Ray remote function, the type of `results` is `RayStorage`,
    when operand is executed, it will fetch dependencies from a Ray actor.
    """

    _graph_execution_cls = GraphExecutionForRay

    @classmethod
    def handle(cls, op, results, mock=False):
        method_name, mapper = ('execute', cls._op_runners) if not mock else \
            ('estimate_size', cls._op_size_estimators)
        try:
            runner = mapper[type(op)]
        except KeyError:
            runner = getattr(op, method_name)

        # register a custom serializer for Mars operand
        _register_ray_serializer(op)

        try:
            ray.wait([execute_on_ray.remote(runner, results, op)])
            return
        except NotImplementedError:
            for op_cls in mapper.keys():
                if isinstance(op, op_cls):
                    mapper[type(op)] = mapper[op_cls]
                    runner = mapper[op_cls]

                    ray.wait(
                        [execute_on_ray.remote(runner, results, op)])
                    return
            raise KeyError(f'No handler found for op: {op}')

    @classmethod
    def _get_chunk_shape(cls, chunk_key, chunk_result):
        assert isinstance(chunk_result, RayStorage)
        return ray.get(chunk_result.meta_store.get_shape.remote(chunk_key))


class RaySession:
    """
    Session to submit Mars job to Ray cluster.

    If Ray is not initialized, kwargs will pass to initialize Ray.
    """
    def __init__(self, **kwargs):
        # as we cannot serialize fuse chunk for now,
        # we just disable numexpr for ray executor
        engine = kwargs.pop('engine', ['numpy', 'dataframe'])
        if not ray.is_initialized():
            ray.init(**kwargs)
        self._session_id = uuid.uuid4()
        self._executor = RayExecutor(engine=engine,
                                     storage=RayStorage())

    @property
    def session_id(self):
        return self._session_id

    @property
    def executor(self):
        return self._executor

    def fetch(self, *tileables, **kw):
        return self._executor.fetch_tileables(tileables, **kw)

    def fetch_log(self, tileables, offsets=None, sizes=None):  # pragma: no cover
        raise NotImplementedError('`fetch_log` is not implemented for ray executor')

    def run(self, *tileables, **kw):
        """
        Parallelism equals to Ray cluster CPUs.
        """
        if 'n_parallel' not in kw:  # pragma: no cover
            kw['n_parallel'] = ray.cluster_resources()['CPU']
        return self._executor.execute_tileables(tileables, **kw)

    def _update_tileable_shape(self, tileable):
        from ..optimizes.tileable_graph import tileable_optimized

        new_nsplits = self._executor.get_tileable_nsplits(tileable)
        tiled = get_tiled(tileable, mapping=tileable_optimized)
        for t in (tileable, tiled):
            t._update_shape(tuple(sum(nsplit) for nsplit in new_nsplits))
        tiled.nsplits = new_nsplits

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._executor = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import BoolField, AnyField, StringField
from ..core import IndexValue
from ..operands import DataFrameOperandMixin, DataFrameOperand, DATAFRAME_TYPE
from ..utils import parse_index, build_empty_df, build_empty_series, standardize_range_index


class DataFrameResetIndex(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.RESET_INDEX

    _level = AnyField('level')
    _drop = BoolField('drop')
    _name = StringField('name')
    _col_level = AnyField('col_level')
    _col_fill = AnyField('col_fill')
    _incremental_index = BoolField('incremental_index')

    def __init__(self, level=None, drop=None, name=None, col_level=None, col_fill=None,
                 incremental_index=None, output_types=None, **kwargs):
        super().__init__(_level=level, _drop=drop, _name=name, _col_level=col_level,
                         _col_fill=col_fill, _incremental_index=incremental_index,
                         _output_types=output_types, **kwargs)

    @property
    def level(self):
        return self._level

    @property
    def drop(self):
        return self._drop

    @property
    def name(self):
        return self._name

    @property
    def col_level(self):
        return self._col_level

    @property
    def col_fill(self):
        return self._col_fill

    @property
    def incremental_index(self):
        return self._incremental_index

    @classmethod
    def _tile_series(cls, op: "DataFrameResetIndex"):
        out_chunks = []
        out = op.outputs[0]
        is_range_index = out.index_value.has_value()
        cum_range = np.cumsum((0, ) + op.inputs[0].nsplits[0])
        for c in op.inputs[0].chunks:
            if is_range_index:
                index_value = parse_index(pd.RangeIndex(cum_range[c.index[0]], cum_range[c.index[0] + 1]))
            else:
                index_value = out.index_value
            chunk_op = op.copy().reset_key()
            if op.drop:
                out_chunk = chunk_op.new_chunk([c], shape=c.shape, index=c.index, dtype=c.dtype,
                                               name=c.name, index_value=index_value)
            else:
                shape = (c.shape[0], out.shape[1])
                out_chunk = chunk_op.new_chunk([c], shape=shape, index=c.index + (0,), dtypes=out.dtypes,
                                               index_value=index_value, columns_value=out.columns_value)
            out_chunks.append(out_chunk)
        if not is_range_index and isinstance(out.index_value.value, IndexValue.RangeIndex) and \
                op.incremental_index:
            out_chunks = standardize_range_index(out_chunks)
        new_op = op.copy()
        if op.drop:
            return new_op.new_seriess(op.inputs, op.inputs[0].shape, nsplits=op.inputs[0].nsplits,
                                      name=out.name, chunks=out_chunks, dtype=out.dtype,
                                      index_value=out.index_value)
        else:
            nsplits = (op.inputs[0].nsplits[0], (out.shape[1],))
            return new_op.new_dataframes(op.inputs, out.shape, nsplits=nsplits, chunks=out_chunks,
                                         index_value=out.index_value, columns_value=out.columns_value,
                                         dtypes=out.dtypes)

    @classmethod
    def _tile_dataframe(cls, op: "DataFrameResetIndex"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        added_columns_num = len(out_df.dtypes) - len(in_df.dtypes)
        out_chunks = []
        index_has_value = out_df.index_value.has_value()
        chunk_has_nan = any(np.isnan(s) for s in in_df.nsplits[0])
        cum_range = np.cumsum((0, ) + in_df.nsplits[0])
        for c in in_df.chunks:
            if index_has_value:
                if chunk_has_nan:
                    index_value = parse_index(pd.RangeIndex(-1))
                else:
                    index_value = parse_index(pd.RangeIndex(cum_range[c.index[0]], cum_range[c.index[0] + 1]))
            else:
                index_value = out_df.index_value
            if c.index[1] == 0:
                chunk_op = op.copy().reset_key()
                dtypes = out_df.dtypes[:(added_columns_num + len(c.dtypes))]
                columns_value = parse_index(dtypes.index)
                new_chunk = chunk_op.new_chunk([c], shape=(c.shape[0], c.shape[1] + added_columns_num),
                                               index=c.index, index_value=index_value,
                                               columns_value=columns_value, dtypes=dtypes)
            else:
                chunk_op = op.copy().reset_key()
                chunk_op._drop = True
                new_chunk = chunk_op.new_chunk([c], shape=c.shape, index_value=index_value,
                                               index=c.index, columns_value=c.columns_value, dtypes=c.dtypes)
            out_chunks.append(new_chunk)
        if not index_has_value or chunk_has_nan:
            if isinstance(out_df.index_value.value, IndexValue.RangeIndex) and op.incremental_index:
                out_chunks = standardize_range_index(out_chunks)
        new_op = op.copy()
        columns_splits = list(in_df.nsplits[1])
        columns_splits[0] += added_columns_num
        nsplits = (in_df.nsplits[0], tuple(columns_splits))
        return new_op.new_dataframes(op.inputs, out_df.shape, nsplits=nsplits,
                                     chunks=out_chunks, dtypes=out_df.dtypes,
                                     index_value=out_df.index_value, columns_value=out_df.columns_value)

    @classmethod
    def tile(cls, op):
        if isinstance(op.inputs[0], DATAFRAME_TYPE):
            return cls._tile_dataframe(op)
        else:
            return cls._tile_series(op)

    @classmethod
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        out = op.outputs[0]

        kwargs = dict()
        if op.name is not None:
            kwargs['name'] = op.name
        if op.col_level is not None:
            kwargs['col_level'] = op.col_level
        if op.col_fill is not None:
            kwargs['col_fill'] = op.col_fill

        r = in_data.reset_index(level=op.level, drop=op.drop, **kwargs)
        if out.index_value.has_value():
            r.index = out.index_value.to_pandas()
        ctx[out.key] = r

    @classmethod
    def _get_out_index(cls, df, out_shape):
        if isinstance(df.index, pd.RangeIndex):
            range_value = -1 if np.isnan(out_shape[0]) else out_shape[0]
            index_value = parse_index(pd.RangeIndex(range_value))
        else:
            index_value = parse_index(df.index)
        return index_value

    def _call_series(self, a):
        if self.drop:
            range_value = -1 if np.isnan(a.shape[0]) else a.shape[0]
            index_value = parse_index(pd.RangeIndex(range_value))
            return self.new_series([a], shape=a.shape, dtype=a.dtype, name=a.name, index_value=index_value)
        else:
            empty_series = build_empty_series(dtype=a.dtype, index=a.index_value.to_pandas()[:0], name=a.name)
            empty_df = empty_series.reset_index(level=self.level, name=self.name)
            shape = (a.shape[0], len(empty_df.dtypes))
            index_value = self._get_out_index(empty_df, shape)
            return self.new_dataframe([a], shape=shape, index_value=index_value,
                                      columns_value=parse_index(empty_df.columns),
                                      dtypes=empty_df.dtypes)

    def _call_dataframe(self, a):
        if self.drop:
            shape = a.shape
            columns_value = a.columns_value
            dtypes = a.dtypes
            range_value = -1 if np.isnan(a.shape[0]) else a.shape[0]
            index_value = parse_index(pd.RangeIndex(range_value))
        else:
            empty_df = build_empty_df(a.dtypes)
            empty_df.index = a.index_value.to_pandas()[:0]
            empty_df = empty_df.reset_index(level=self.level, col_level=self.col_level, col_fill=self.col_fill)
            shape = (a.shape[0], len(empty_df.columns))
            columns_value = parse_index(empty_df.columns, store_data=True)
            dtypes = empty_df.dtypes
            index_value = self._get_out_index(empty_df, shape)
        return self.new_dataframe([a], shape=shape, columns_value=columns_value,
                                  index_value=index_value, dtypes=dtypes)

    def __call__(self, a):
        if isinstance(a, DATAFRAME_TYPE):
            return self._call_dataframe(a)
        else:
            return self._call_series(a)


def df_reset_index(df, level=None, drop=False, inplace=False, col_level=0,
                   col_fill='', incremental_index=False):
    """
    Reset the index, or a level of it.

    Reset the index of the DataFrame, and use the default one instead.
    If the DataFrame has a MultiIndex, this method can remove one or more
    levels.

    Parameters
    ----------
    level : int, str, tuple, or list, default None
        Only remove the given levels from the index. Removes all levels by
        default.
    drop : bool, default False
        Do not try to insert index into dataframe columns. This resets
        the index to the default integer index.
    inplace : bool, default False
        Modify the DataFrame in place (do not create a new object).
    col_level : int or str, default 0
        If the columns have multiple levels, determines which level the
        labels are inserted into. By default it is inserted into the first
        level.
    col_fill : object, default ''
        If the columns have multiple levels, determines how the other
        levels are named. If None then the index name is repeated.
    incremental_index: bool, default False
        Ensure RangeIndex incremental, when output DataFrame has multiple chunks,
        ensuring index incremental costs more computation,
        so by default, each chunk will have index which starts from 0,
        setting incremental_index=True，reset_index will garantee that
        output DataFrame's index is from 0 to n - 1.

    Returns
    -------
    DataFrame or None
        DataFrame with the new index or None if ``inplace=True``.

    See Also
    --------
    DataFrame.set_index : Opposite of reset_index.
    DataFrame.reindex : Change to new indices or expand indices.
    DataFrame.reindex_like : Change to same indices as other DataFrame.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> df = md.DataFrame([('bird', 389.0),
    ...                    ('bird', 24.0),
    ...                    ('mammal', 80.5),
    ...                    ('mammal', mt.nan)],
    ...                   index=['falcon', 'parrot', 'lion', 'monkey'],
    ...                   columns=('class', 'max_speed'))
    >>> df.execute()
             class  max_speed
    falcon    bird      389.0
    parrot    bird       24.0
    lion    mammal       80.5
    monkey  mammal        NaN

    When we reset the index, the old index is added as a column, and a
    new sequential index is used:

    >>> df.reset_index().execute()
        index   class  max_speed
    0  falcon    bird      389.0
    1  parrot    bird       24.0
    2    lion  mammal       80.5
    3  monkey  mammal        NaN

    We can use the `drop` parameter to avoid the old index being added as
    a column:

    >>> df.reset_index(drop=True).execute()
        class  max_speed
    0    bird      389.0
    1    bird       24.0
    2  mammal       80.5
    3  mammal        NaN

    You can also use `reset_index` with `MultiIndex`.

    >>> import pandas as pd
    >>> index = pd.MultiIndex.from_tuples([('bird', 'falcon'),
    ...                                    ('bird', 'parrot'),
    ...                                    ('mammal', 'lion'),
    ...                                    ('mammal', 'monkey')],
    ...                                   names=['class', 'name'])
    >>> columns = pd.MultiIndex.from_tuples([('speed', 'max'),
    ...                                      ('species', 'type')])
    >>> df = md.DataFrame([(389.0, 'fly'),
    ...                    ( 24.0, 'fly'),
    ...                    ( 80.5, 'run'),
    ...                    (mt.nan, 'jump')],
    ...                   index=index,
    ...                   columns=columns)
    >>> df.execute()
                   speed species
                     max    type
    class  name
    bird   falcon  389.0     fly
           parrot   24.0     fly
    mammal lion     80.5     run
           monkey    NaN    jump

    If the index has multiple levels, we can reset a subset of them:

    >>> df.reset_index(level='class').execute()
             class  speed species
                      max    type
    name
    falcon    bird  389.0     fly
    parrot    bird   24.0     fly
    lion    mammal   80.5     run
    monkey  mammal    NaN    jump

    If we are not dropping the index, by default, it is placed in the top
    level. We can place it in another level:

    >>> df.reset_index(level='class', col_level=1).execute()
                    speed species
             class    max    type
    name
    falcon    bird  389.0     fly
    parrot    bird   24.0     fly
    lion    mammal   80.5     run
    monkey  mammal    NaN    jump

    When the index is inserted under another level, we can specify under
    which one with the parameter `col_fill`:

    >>> df.reset_index(level='class', col_level=1, col_fill='species').execute()
                  species  speed species
                    class    max    type
    name
    falcon           bird  389.0     fly
    parrot           bird   24.0     fly
    lion           mammal   80.5     run
    monkey         mammal    NaN    jump

    If we specify a nonexistent level for `col_fill`, it is created:

    >>> df.reset_index(level='class', col_level=1, col_fill='genus').execute()
                    genus  speed species
                    class    max    type
    name
    falcon           bird  389.0     fly
    parrot           bird   24.0     fly
    lion           mammal   80.5     run
    monkey         mammal    NaN    jump
    """
    op = DataFrameResetIndex(level=level, drop=drop, col_level=col_level,
                             col_fill=col_fill, incremental_index=incremental_index,
                             output_types=[OutputType.dataframe])
    ret = op(df)
    if not inplace:
        return ret
    else:
        df.data = ret.data


def series_reset_index(series, level=None, drop=False, name=None,
                       inplace=False, incremental_index=False):
    """
    Generate a new DataFrame or Series with the index reset.

    This is useful when the index needs to be treated as a column, or
    when the index is meaningless and needs to be reset to the default
    before another operation.

    Parameters
    ----------
    level : int, str, tuple, or list, default optional
        For a Series with a MultiIndex, only remove the specified levels
        from the index. Removes all levels by default.
    drop : bool, default False
        Just reset the index, without inserting it as a column in
        the new DataFrame.
    name : object, optional
        The name to use for the column containing the original Series
        values. Uses ``self.name`` by default. This argument is ignored
        when `drop` is True.
    inplace : bool, default False
        Modify the Series in place (do not create a new object).
    incremental_index: bool, default False
        Ensure RangeIndex incremental, when output Series has multiple chunks,
        ensuring index incremental costs more computation,
        so by default, each chunk will have index which starts from 0,
        setting incremental_index=True，reset_index will garantee that
        output Series's index is from 0 to n - 1.

    Returns
    -------
    Series or DataFrame
        When `drop` is False (the default), a DataFrame is returned.
        The newly created columns will come first in the DataFrame,
        followed by the original Series values.
        When `drop` is True, a `Series` is returned.
        In either case, if ``inplace=True``, no value is returned.

    See Also
    --------
    DataFrame.reset_index: Analogous function for DataFrame.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 2, 3, 4], name='foo',
    ...               index=md.Index(['a', 'b', 'c', 'd'], name='idx'))

    Generate a DataFrame with default index.

    >>> s.reset_index().execute()
      idx  foo
    0   a    1
    1   b    2
    2   c    3
    3   d    4

    To specify the name of the new column use `name`.

    >>> s.reset_index(name='values').execute()
      idx  values
    0   a       1
    1   b       2
    2   c       3
    3   d       4

    To generate a new Series with the default set `drop` to True.

    >>> s.reset_index(drop=True).execute()
    0    1
    1    2
    2    3
    3    4
    Name: foo, dtype: int64

    To update the Series in place, without generating a new one
    set `inplace` to True. Note that it also requires ``drop=True``.

    >>> s.reset_index(inplace=True, drop=True)
    >>> s.execute()
    0    1
    1    2
    2    3
    3    4
    Name: foo, dtype: int64

    The `level` parameter is interesting for Series with a multi-level
    index.

    >>> import numpy as np
    >>> import pandas as pd
    >>> arrays = [np.array(['bar', 'bar', 'baz', 'baz']),
    ...           np.array(['one', 'two', 'one', 'two'])]
    >>> s2 = md.Series(
    ...     range(4), name='foo',
    ...     index=pd.MultiIndex.from_arrays(arrays,
    ...                                     names=['a', 'b']))

    To remove a specific level from the Index, use `level`.

    >>> s2.reset_index(level='a').execute()
           a  foo
    b
    one  bar    0
    two  bar    1
    one  baz    2
    two  baz    3

    If `level` is not set, all levels are removed from the Index.

    >>> s2.reset_index().execute()
         a    b  foo
    0  bar  one    0
    1  bar  two    1
    2  baz  one    2
    3  baz  two    3
    """
    op = DataFrameResetIndex(level=level, drop=drop, name=name,
                             incremental_index=incremental_index,
                             output_types=[OutputType.series])
    ret = op(series)
    if not inplace:
        return ret
    elif ret.ndim == 2:
        raise TypeError('Cannot reset_index inplace on a Series to create a DataFrame')
    else:
        series.data = ret.data



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from distutils.version import LooseVersion

import numpy as np

from ...core import ExecutableTuple
from ..arithmetic import (
    divide as mt_divide, sqrt as mt_sqrt, absolute as mt_abs,
    isnan as mt_isnan,
)
from ..base import where as mt_where
from ..reduction import (
    var as mt_var, mean as mt_mean,
)
from ..utils import implement_scipy

try:
    from scipy import __version__ as sp_version
    from scipy.stats import (
        ttest_ind as sp_ttest_ind,
        ttest_ind_from_stats as sp_ttest_ind_from_stats,
        ttest_rel as sp_ttest_rel,
        ttest_1samp as sp_ttest_1samp,
    )
    from scipy.stats import distributions as sp_distributions
except ImportError:
    sp_version = None
    sp_ttest_1samp = sp_ttest_ind = sp_ttest_ind_from_stats = sp_ttest_rel = None
    sp_distributions = None


def _equal_var_ttest_denom(v1, n1, v2, n2):
    df = n1 + n2 - 2.0
    svar = ((n1 - 1) * v1 + (n2 - 1) * v2) / df
    denom = mt_sqrt(svar * (1.0 / n1 + 1.0 / n2))  # XXX: np -> da
    return df, denom


def _unequal_var_ttest_denom(v1, n1, v2, n2):
    vn1 = v1 / n1
    vn2 = v2 / n2
    with np.errstate(divide="ignore", invalid="ignore"):
        df = (vn1 + vn2) ** 2 / (vn1 ** 2 / (n1 - 1) + vn2 ** 2 / (n2 - 1))

    # If df is undefined, variances are zero (assumes n1 > 0 & n2 > 0).
    # Hence it doesn't matter what df is as long as it's not NaN.
    df = mt_where(mt_isnan(df), 1, df)
    denom = mt_sqrt(vn1 + vn2)
    return df, denom


def _ttest_ind_from_stats(mean1, mean2, denom, df, alternative):

    d = mean1 - mean2
    with np.errstate(divide="ignore", invalid="ignore"):
        t = mt_divide(d, denom)
    t, prob = _ttest_finish(df, t, alternative)

    return t, prob


def _ttest_finish(df, t, alternative):
    """Common code between all 3 t-test functions."""
    if alternative != 'two-sided' and LooseVersion(sp_version) < '1.6.0':  # pragma: no cover
        raise ValueError("alternative must be 'two-sided' with scipy prior to 1.6.0")

    if alternative == 'less':
        prob = t.map_chunk(sp_distributions.t.cdf, args=(df,))
    elif alternative == 'greater':
        prob = t.map_chunk(sp_distributions.t.sf, args=(df,))
    elif alternative == 'two-sided':
        prob = mt_abs(t).map_chunk(sp_distributions.t.sf, args=(df,)) * 2
    else:
        raise ValueError("alternative must be "
                         "'less', 'greater' or 'two-sided'")
    if t.ndim == 0:
        t = t[()]
    return t, prob


Ttest_1sampResult = namedtuple('Ttest_1sampResult', ('statistic', 'pvalue'))


@implement_scipy(sp_ttest_1samp)
def ttest_1samp(a, popmean, axis=0, nan_policy="propagate", alternative="two-sided"):
    if nan_policy != "propagate":
        raise NotImplementedError(
            "`nan_policy` other than 'propagate' have not been implemented."
        )
    n = a.shape[axis]
    df = n - 1

    d = a.mean(axis=axis) - popmean
    v = a.var(axis=axis, ddof=1)
    denom = mt_sqrt(v / float(n))

    with np.errstate(divide="ignore", invalid="ignore"):
        t = mt_divide(d, denom)
    t, prob = _ttest_finish(df, t, alternative)
    return ExecutableTuple(Ttest_1sampResult(t, prob))


Ttest_indResult = namedtuple('Ttest_indResult', ('statistic', 'pvalue'))


@implement_scipy(sp_ttest_ind)
def ttest_ind(a, b, axis=0, equal_var=True, alternative="two-sided"):
    v1 = mt_var(a, axis, ddof=1)
    v2 = mt_var(b, axis, ddof=1)
    n1 = a.shape[axis]
    n2 = b.shape[axis]

    if equal_var:
        df, denom = _equal_var_ttest_denom(v1, n1, v2, n2)
    else:
        df, denom = _unequal_var_ttest_denom(v1, n1, v2, n2)

    res = _ttest_ind_from_stats(mt_mean(a, axis), mt_mean(b, axis), denom,
                                df, alternative)

    return ExecutableTuple(Ttest_indResult(*res))


@implement_scipy(sp_ttest_ind_from_stats)
def ttest_ind_from_stats(mean1, std1, nobs1, mean2, std2, nobs2,
                         equal_var=True, alternative="two-sided"):
    if equal_var:
        df, denom = _equal_var_ttest_denom(std1**2, nobs1, std2**2, nobs2)
    else:
        df, denom = _unequal_var_ttest_denom(std1**2, nobs1,
                                             std2**2, nobs2)

    res = _ttest_ind_from_stats(mean1, mean2, denom, df, alternative)
    return ExecutableTuple(Ttest_indResult(*res))


Ttest_relResult = namedtuple('Ttest_relResult', ('statistic', 'pvalue'))


@implement_scipy(sp_ttest_rel)
def ttest_rel(a, b, axis=0, nan_policy="propagate", alternative="two-sided"):
    if nan_policy != "propagate":
        raise NotImplementedError(
            "`nan_policy` other than 'propagate' have not been implemented."
        )

    n = a.shape[axis]
    df = float(n - 1)

    d = (a - b).astype(np.float64)
    v = mt_var(d, axis, ddof=1)
    dm = mt_mean(d, axis)
    denom = mt_sqrt(v / float(n))

    with np.errstate(divide="ignore", invalid="ignore"):
        t = mt_divide(dm, denom)
    t, prob = _ttest_finish(df, t, alternative)

    return ExecutableTuple(Ttest_relResult(t, prob))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import infer_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorInvert(TensorUnaryOp):
    _op_type_ = OperandDef.INVERT
    _func_name = 'invert'


@infer_dtype(np.invert)
def invert(x, out=None, where=None, **kwargs):
    """
    Compute bit-wise inversion, or bit-wise NOT, element-wise.

    Computes the bit-wise NOT of the underlying binary representation of
    the integers in the input tensors. This ufunc implements the C/Python
    operator ``~``.

    For signed integer inputs, the two's complement is returned.  In a
    two's-complement system negative numbers are represented by the two's
    complement of the absolute value. This is the most common method of
    representing signed integers on computers [1]_. A N-bit
    two's-complement system can represent every integer in the range
    :math:`-2^{N-1}` to :math:`+2^{N-1}-1`.

    Parameters
    ----------
    x : array_like
        Only integer and boolean types are handled.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : array_like
        Result.

    See Also
    --------
    bitwise_and, bitwise_or, bitwise_xor
    logical_not

    Notes
    -----
    `bitwise_not` is an alias for `invert`:

    >>> import mars.tensor as mt

    >>> mt.bitwise_not is mt.invert
    True

    References
    ----------
    .. [1] Wikipedia, "Two's complement",
        http://en.wikipedia.org/wiki/Two's_complement

    Examples
    --------
    We've seen that 13 is represented by ``00001101``.
    The invert or bit-wise NOT of 13 is then:

    >>> mt.invert(mt.array([13], dtype=mt.uint8)).execute()
    array([242], dtype=uint8)

    The result depends on the bit-width:

    >>> mt.invert(mt.array([13], dtype=mt.uint16)).execute()
    array([65522], dtype=uint16)

    When using signed integer types the result is the two's complement of
    the result for the unsigned type:

    >>> mt.invert(mt.array([13], dtype=mt.int8)).execute()
    array([-14], dtype=int8)

    Booleans are accepted as well:

    >>> mt.invert(mt.array([True, False])).execute()
    array([False,  True])
    """
    op = TensorInvert(**kwargs)
    return op(x, out=out, where=where)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..utils import inject_dtype
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode='unary')
class TensorIsComplex(TensorUnaryOp):
    _op_type_ = OperandDef.ISCOMPLEX
    _func_name = 'iscomplex'


@inject_dtype(np.bool_)
def iscomplex(x, **kwargs):
    """
    Returns a bool tensor, where True if input element is complex.

    What is tested is whether the input has a non-zero imaginary part, not if
    the input type is complex.

    Parameters
    ----------
    x : array_like
        Input tensor.

    Returns
    -------
    out : Tensor of bools
        Output tensor.

    See Also
    --------
    isreal
    iscomplexobj : Return True if x is a complex type or an array of complex
                   numbers.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.iscomplex([1+1j, 1+0j, 4.5, 3, 2, 2j]).execute()
    array([ True, False, False, False, False,  True])

    """
    op = TensorIsComplex(**kwargs)
    return op(x)
